from io import StringIO
from json import dumps as json_dumps
import pathlib
import sys
from typing import Any, Dict, List, Optional, Tuple

import click
from rich.console import Console
from rich.table import Table
import yaml

from anyscale._private.models.image_uri import ImageURI
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.list_util import (
    display_list,
    MAX_PAGE_SIZE,
    NON_INTERACTIVE_DEFAULT_MAX_ITEMS,
    validate_page_size,
)
from anyscale.commands.util import (
    AnyscaleCommand,
    build_kv_table,
    convert_kv_strings_to_dict,
    override_env_vars,
    parse_repeatable_tags_to_dict,
    parse_tags_kv_to_str_map,
)
from anyscale.controllers.service_controller import ServiceController
import anyscale.service
from anyscale.service.models import (
    ServiceConfig,
    ServiceLogMode,
    ServiceSortField,
    ServiceSortOrder,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
)
from anyscale.util import (
    AnyscaleJSONEncoder,
    get_endpoint,
    validate_non_negative_arg,
    validate_service_state_filter,
)


log = BlockLogger()  # CLI Logger


@click.group("service")
def service_cli():
    pass


def _read_name_from_config_file(path: str):
    """Read the 'name' property from the config file at `path`.

    This enables reading the name for both the existing (rollout) and new (deploy)
    config file formats.
    """
    if not pathlib.Path(path).is_file():
        raise click.ClickException(f"Config file not found at path: '{path}'.")

    with open(path) as f:
        config = yaml.safe_load(f)

    if config is None or "name" not in config:
        raise click.ClickException(f"No 'name' property found in config file '{path}'.")

    return config["name"]


@service_cli.command(
    name="deploy", help="Deploy or update a service.",
)
@click.argument("import_path", type=str, required=False, default=None)
@click.argument("arguments", nargs=-1, required=False)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=[],
    type=str,
    multiple=True,
    help="Path to a YAML config file to deploy. When deploying from a file, import path and arguments cannot be provided. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "-n",
    "--name",
    required=False,
    default=None,
    type=str,
    help="Unique name for the service. When running in a workspace, this defaults to the workspace name.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the service. This cannot be changed when using --in-place and is exclusive with --containerfile. When running in a workspace, this defaults to the image of the workspace.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the service. This cannot be changed when using --in-place and is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If you don't specify a Ray version, Anyscale defaults to the latest Ray version available at the time of the Anyscale CLI/SDK release.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the service. This cannot be changed when using --in-place. When running in a workspace, this defaults to the compute configuration of the workspace.",
)
@click.option(
    "-w",
    "--working-dir",
    required=False,
    default=None,
    type=str,
    help="Path to a local directory or a remote URI to a .zip file (S3, GS, HTTP) that will be the working directory for the service. The files in the directory will be automatically uploaded to cloud storage. When running in a workspace, this defaults to the current working directory.",
)
@click.option(
    "-e",
    "--exclude",
    required=False,
    type=str,
    multiple=True,
    help="File pattern to exclude when uploading local directories. This argument can be specified multiple times and the patterns will be appended to the 'excludes' list in the config file (if any).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the service. Anyscale installs these dependencies on top of the image. If you deploy a service from a workspace, the default is to use the workspace dependencies, but specifying this option overrides them.",
)
@click.option(
    "-i",
    "--in-place",
    is_flag=True,
    show_default=True,
    default=False,
    help="Perform an in-place upgrade without starting a new cluster. This can be used for faster iteration during development but is *not* currently recommended for production deploys. This *cannot* be used to change cluster-level options such as image and compute config (they will be ignored).",
)
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic to send to the canary version of the service (0-100). This can be used to manually shift traffic toward (or away from) the canary version. If not provided, traffic will be shifted incrementally toward the canary version until it reaches 100. Not supported when using --in-place. This is ignored when restarting a service or creating a new service.",
)
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Amount of excess capacity allowed to be used while updating the service (0-100). Defaults to 100. Not supported when using --in-place.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the service. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
@click.option(
    "--py-module",
    required=False,
    default=None,
    multiple=True,
    type=str,
    help="Python modules to be available for import in the Ray workers. Each entry must be a path to a local directory.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--versions",
    required=False,
    default=None,
    type=str,
    help="Defines the traffic and capacity percents per version. Capacity defaults to traffic.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
@click.option(
    "--version-name",
    required=False,
    default=None,
    type=str,
    help="Unique name for the service version. This can only be used for single version deployments. For multi-version deployments, specify version names in the config files and --versions.",
)
def deploy(  # noqa: PLR0912, PLR0913 C901
    config_file: List[str],
    import_path: Optional[str],
    arguments: Tuple[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    working_dir: Optional[str],
    exclude: Tuple[str],
    requirements: Optional[str],
    in_place: bool,
    canary_percent: Optional[int],
    max_surge_percent: Optional[int],
    env: Optional[Tuple[str]],
    py_module: Tuple[str],
    cloud: Optional[str],
    project: Optional[str],
    versions: Optional[str],
    tags: Optional[Tuple[str]],
    version_name: Optional[str],
):
    """Deploy or update a service.

    If no service with the provided name is running, one will be created, else the existing service will be updated.

    To deploy a single application, the import path and optional arguments can be passed directly as arguments:

    `$ anyscale service deploy main:app arg1=val1`

    To deploy multiple applications, deploy from a config file using `-f`:

    `$ anyscale service deploy -f config.yaml`

    Command-line flags override values in the config file.
    """

    if versions is None:
        if len(config_file) == 1:
            if import_path is not None or len(arguments) > 0:
                raise click.ClickException(
                    "When a config file is provided, import path and application arguments can't be."
                )

            if not pathlib.Path(config_file[0]).is_file():
                raise click.ClickException(f"Config file '{config_file[0]}' not found.")

            config = ServiceConfig.from_yaml(config_file[0])
        elif len(config_file) > 1:
            raise click.ClickException(
                "Multiple config files can be provided only when deploying multiple versions with --versions."
            )
        else:
            # when config_file is not provided.
            if import_path is None:
                raise click.ClickException(
                    "Either config file or import path must be provided."
                )

            if (
                import_path.endswith((".yaml", ".yml"))
                or pathlib.Path(import_path).is_file()
            ):
                log.warning(
                    f"The provided import path '{import_path}' looks like a config file. Did you mean to use '-f config.yaml'?"
                )

            app: Dict[str, Any] = {"import_path": import_path}
            arguments_dict = convert_kv_strings_to_dict(arguments)
            if arguments_dict:
                app["args"] = arguments_dict

            config = ServiceConfig(applications=[app])

        if containerfile and image_uri:
            raise click.ClickException(
                "Only one of '--containerfile' and '--image-uri' can be provided."
            )

        if ray_version and (not image_uri and not containerfile):
            raise click.ClickException(
                "Ray version can only be used with an image or containerfile.",
            )

        if registry_login_secret and (
            not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
        ):
            raise click.ClickException(
                "Registry login secret can only be used with an image that is not hosted on Anyscale."
            )

        if name is not None:
            config = config.options(name=name)

        if image_uri is not None:
            config = config.options(image_uri=image_uri)

        if registry_login_secret is not None:
            config = config.options(registry_login_secret=registry_login_secret)

        if ray_version is not None:
            config = config.options(ray_version=ray_version)

        if containerfile is not None:
            config = config.options(containerfile=containerfile)

        if compute_config is not None:
            config = config.options(compute_config=compute_config)

        if working_dir is not None:
            config = config.options(working_dir=working_dir)

        if exclude:
            config = config.options(excludes=[e for e in exclude])

        if requirements is not None:
            config = config.options(requirements=requirements)

        if env:
            config = override_env_vars(config, convert_kv_strings_to_dict(env))

        if py_module:
            for module in py_module:
                if not pathlib.Path(module).is_dir():
                    raise click.ClickException(
                        f"Python module path '{module}' does not exist or is not a directory."
                    )
            config = config.options(py_modules=[*py_module])

        if cloud is not None:
            config = config.options(cloud=cloud)
        if project is not None:
            config = config.options(project=project)

        if version_name is not None:
            config = config.options(version_name=version_name)

        configs = config
    else:
        # When multiple versions are being deployed.
        configs = []
        for config_path in config_file:
            svc_config = ServiceConfig.from_yaml(config_path)
            if name is not None:
                svc_config = svc_config.options(name=name)

            if cloud is not None:
                svc_config = svc_config.options(cloud=cloud)

            if project is not None:
                svc_config = svc_config.options(project=project)

            if image_uri is not None:
                log.warning("--image-uri is ignored.")

            if registry_login_secret is not None:
                log.warning("--registry-login-secret is ignored.")

            if ray_version is not None:
                log.warning("--ray-version is ignored.")

            if containerfile is not None:
                log.warning("--containerfile is ignored.")

            if compute_config is not None:
                log.warning("--compute-config is ignored.")

            if working_dir is not None:
                log.warning("--working-dir is ignored.")

            if exclude:
                log.warning("--exclude is ignored.")

            if requirements is not None:
                log.warning("--requirements is ignored.")

            if env:
                log.warning("--env is ignored.")

            if py_module:
                log.warning("--py-module is ignored.")

            if version_name is not None:
                log.warning("--version-name is ignored.")

            configs.append(svc_config)

    if tags:
        tag_map = parse_tags_kv_to_str_map(tags)
        if tag_map:
            if isinstance(configs, ServiceConfig):
                configs = configs.options(tags=tag_map)
            else:
                configs = [cfg.options(tags=tag_map) for cfg in configs]

    anyscale.service.deploy(
        configs,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
        versions=versions,
        name=name,
        cloud=cloud,
        project=project,
    )


@service_cli.command(
    name="status", help="Get the status of a service.",
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to read the name from.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-j",
    "--json",
    is_flag=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Include verbose details in the status.",
)
def status(
    name: Optional[str],
    config_file: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    json: bool,
    verbose: bool,
):
    if name is not None and config_file is not None:
        raise click.ClickException(
            "Only one of '--name' and '--config-file' can be provided."
        )

    if config_file is not None:
        name = _read_name_from_config_file(config_file)

    if name is None:
        raise click.ClickException(
            "Service name must be provided using '--name' or in a config file using '-f'."
        )

    status: ServiceStatus = anyscale.service.status(
        name=name, cloud=cloud, project=project
    )
    status_dict = status.to_dict()
    if not verbose:
        # TODO(edoakes): consider adding this as an API on the model itself if it
        # becomes a common pattern.
        status_dict.get("primary_version", {}).pop("config", None)
        status_dict.get("canary_version", {}).pop("config", None)
        # Remove config from all versions in multi-version services
        if status_dict.get("versions"):
            for version in status_dict.get("versions", []):
                version.pop("config", None)

    console = Console()
    if json:
        json_str = json_dumps(status_dict, indent=2, cls=AnyscaleJSONEncoder)
        console.print_json(json=json_str)
    else:
        stream = StringIO()
        yaml.dump(status_dict, stream, sort_keys=False)
        console.print(stream.getvalue(), end="")


@service_cli.command(
    name="wait", help="Wait for a service to enter a target state.",
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to read the name from.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-s",
    "--state",
    default=str(ServiceState.RUNNING),
    help="The ServiceState to wait for the service to reach. Defaults to RUNNING.",
)
@click.option(
    "-t",
    "--timeout-s",
    default=600,
    type=float,
    help="Timeout to wait for the service to reach the target state. Defaults to 600s (10min).",
)
def wait(
    name: Optional[str],
    config_file: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    state: str,
    timeout_s: int,
):
    try:
        state = ServiceState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))

    if name is not None and config_file is not None:
        raise click.ClickException(
            "Only one of '--name' and '--config-file' can be provided."
        )

    if config_file is not None:
        name = _read_name_from_config_file(config_file)

    if name is None:
        raise click.ClickException(
            "Service name must be provided using '--name' or in a config file using '-f'."
        )

    try:
        anyscale.service.wait(
            name=name, cloud=cloud, project=project, state=state, timeout_s=timeout_s
        )
    except Exception as e:  # noqa: BLE001
        raise click.ClickException(str(e)) from None


# This is a private CLI command to be used internally for testing. This is HIDDEN
# from the user and is not documented in the CLI help.
@service_cli.command(
    name="controller-logs", help="View the controller logs of a service.", hidden=True,
)
@click.option(
    "-n", "--name", required=True, type=str, help="Name of the service.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "--canary",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Whether to show the logs of the canary version of the service. If not provided, the primary version logs will be shown.",
)
@click.option(
    "--head",
    required=False,
    default=False,
    type=bool,
    is_flag=True,
    help="Used with --max-lines to get `max-lines` lines from the head of the log.",
)
@click.option(
    "--max-lines",
    required=False,
    default=1000,
    type=int,
    help="Used with --head to limit the number of lines output.",
)
def controller_logs(
    name: str,
    cloud: Optional[str],
    project: Optional[str],
    canary: bool,
    head: bool,
    max_lines: int = 1000,
):
    mode = ServiceLogMode.TAIL
    if head:
        mode = ServiceLogMode.HEAD

    try:
        logs = anyscale.service._controller_logs(  # noqa: SLF001
            name=name,
            cloud=cloud,
            project=project,
            canary=canary,
            mode=mode,
            max_lines=max_lines,
        )
        print(logs)
    except ValueError as e:
        raise click.ClickException(str(e))


def validate_max_items(ctx, param, value):
    if value is None:
        return None
    return validate_non_negative_arg(ctx, param, value)


def _parse_sort_option(sort: Optional[str],) -> Tuple[Optional[str], ServiceSortOrder]:
    """
    Given a raw sort string (e.g. "-created_at"), return
    (canonical_field_name, SortOrder).
    """
    if not sort:
        return None, ServiceSortOrder.ASC

    # build case-insensitive map of allowed fields
    allowed = {f.value.lower(): f.value for f in ServiceSortField.__members__.values()}

    # detect leading '-' for descending
    if sort.startswith("-"):
        raw = sort[1:]
        order = ServiceSortOrder.DESC
    else:
        raw = sort
        order = ServiceSortOrder.ASC

    key = raw.lower()
    if key not in allowed:
        allowed_names = ", ".join(sorted(allowed.values()))
        raise click.BadParameter(
            f"Invalid sort field '{raw}'. Allowed fields: {allowed_names}"
        )

    return allowed[key], order


def _create_service_list_table(show_header: bool) -> Table:
    table = Table(show_header=show_header, expand=True)
    # NAME and ID: larger ratios, can wrap but never truncate
    table.add_column(
        "NAME", no_wrap=False, overflow="fold", ratio=3, min_width=15,
    )
    table.add_column(
        "ID", no_wrap=False, overflow="fold", ratio=2, min_width=12,
    )
    # all other columns will wrap as needed
    for heading in (
        "CURRENT STATE",
        "CREATOR",
        "PROJECT",
        "LAST DEPLOYED AT",
    ):
        table.add_column(
            heading, no_wrap=False, overflow="fold", ratio=1, min_width=8,
        )

    return table


def _format_service_output_data(svc: ServiceStatus) -> Dict[str, str]:
    last_deployed_at = ""
    if isinstance(svc.primary_version, ServiceVersionStatus):
        last_deployed_at = svc.primary_version.created_at.strftime("%Y-%m-%d %H:%M:%S")

    return {
        "name": svc.name,
        "id": svc.id,
        "current_state": str(svc.state),
        "creator": str(svc.creator or ""),
        "project": str(svc.project or ""),
        "last_deployed_at": last_deployed_at,
    }


@service_cli.command(
    name="list",
    help="List services.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_LIST_EXAMPLE,
)
@click.option("--service-id", "--id", help="ID of the service to display.")
@click.option("--name", "-n", help="Name of the service to display.")
@click.option(
    "--cloud",
    type=str,
    help="The Anyscale Cloud of this workload; defaults to your org/workspace cloud.",
)
@click.option(
    "--project",
    type=str,
    help="Named project to use; defaults to your org/workspace project.",
)
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help=(
        "This option can be repeated to filter by multiple tags. "
        "Tags with the same key are ORed, whereas tags with different keys are ANDed. "
        "Example: --tag team:mlops --tag team:infra --tag env:prod. "
        "Filters with team: (mlops OR infra) AND env:prod."
    ),
)
@click.option(
    "--created-by-me",
    is_flag=True,
    default=False,
    help="List services created by me only.",
)
@click.option(
    "--state",
    "-s",
    "state_filter",
    multiple=True,
    callback=validate_service_state_filter,
    help=(
        "Filter by service state (repeatable). "
        f"Allowed: {', '.join(s.value for s in ServiceState)}"
    ),
)
@click.option(
    "--include-archived",
    is_flag=True,
    default=False,
    help="Include archived services.",
)
@click.option(
    "--max-items",
    type=int,
    callback=validate_max_items,
    help="Max total items (only with --no-interactive).",
)
@click.option(
    "--page-size",
    type=int,
    default=10,
    show_default=True,
    callback=validate_page_size,
    help=f"Items per page (max {MAX_PAGE_SIZE}).",
)
@click.option(
    "--interactive/--no-interactive",
    default=True,
    show_default=True,
    help="Use interactive paging.",
)
@click.option(
    "--sort",
    help=(
        "Sort by FIELD (prefix with '-' for desc). "
        f"Allowed: {', '.join(f.value for f in ServiceSortField.__members__.values())}"
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Include full config in JSON output.",
)
@click.option(
    "-j",
    "--json",
    "json_output",
    is_flag=True,
    default=False,
    help="Emit structured JSON to stdout.",
)
def list(  # noqa: PLR0913, A001
    service_id: Optional[str],
    name: Optional[str],
    tags: List[str],
    created_by_me: bool,
    cloud: Optional[str],
    project: Optional[str],
    state_filter: List[str],
    include_archived: bool,
    max_items: Optional[int],
    page_size: int,
    sort: Optional[str],
    json_output: bool,
    interactive: bool,
    verbose: bool,
):
    """List services based on the provided filters."""
    if max_items is not None and interactive:
        raise click.UsageError("--max-items only allowed with --no-interactive")

    # parse sort
    sort_field, sort_order = _parse_sort_option(sort)

    # normalize max_items
    effective_max = max_items
    if not interactive and effective_max is None:
        stderr = Console(stderr=True)
        stderr.print(
            f"Defaulting to {NON_INTERACTIVE_DEFAULT_MAX_ITEMS} items in batch mode; "
            "use --max-items to override."
        )
        effective_max = NON_INTERACTIVE_DEFAULT_MAX_ITEMS

    console = Console()
    stderr = Console(stderr=True)

    # diagnostics
    stderr.print("[bold]Listing services with:[/]")
    stderr.print(f"• name            = {name or '<any>'}")
    stderr.print(f"• states          = {', '.join(state_filter) or '<all>'}")
    stderr.print(f"• tags            = {', '.join(tags) or '<none>'}")
    stderr.print(f"• created_by_me   = {created_by_me}")
    stderr.print(f"• include_archived= {include_archived}")
    stderr.print(f"• sort            = {sort or '<none>'}")
    stderr.print(f"• mode            = {'interactive' if interactive else 'batch'}")
    stderr.print(f"• per-page limit  = {page_size}")
    stderr.print(f"• max-items total = {effective_max or 'all'}")
    stderr.print(f"\nView your Services in the UI at {get_endpoint('/services')}\n")

    creator_id = (
        ServiceController().get_authenticated_user_id() if created_by_me else None
    )

    # choose formatter
    if json_output:

        def json_formatter(svc: ServiceStatus) -> Dict[str, Any]:
            data = svc.to_dict()
            if not verbose:
                data.get("primary_version", {}).pop("config", None)
                data.get("canary_version", {}).pop("config", None)
            return data

        formatter = json_formatter
    else:
        formatter = _format_service_output_data

    total = 0
    try:
        iterator = anyscale.service.list(
            service_id=service_id,
            name=name,
            state_filter=state_filter,
            tags_filter=parse_repeatable_tags_to_dict(tags) if tags else None,
            creator_id=creator_id,
            cloud=cloud,
            project=project,
            include_archived=include_archived,
            max_items=None if interactive else effective_max,
            page_size=page_size,
            sort_field=sort_field,
            sort_order=sort_order,
        )
        total = display_list(
            iterator=iter(iterator),
            item_formatter=formatter,
            table_creator=_create_service_list_table,
            json_output=json_output,
            page_size=page_size,
            interactive=interactive,
            max_items=effective_max,
            console=console,
        )

        if not json_output:
            if total > 0:
                stderr.print(f"\nFetched {total} services.")
            else:
                stderr.print("\nNo services found.")
    except Exception as e:  # noqa: BLE001
        log.error(f"Failed to list services: {e}")
        sys.exit(1)


@service_cli.group("tags", help="Manage tags for services.")
def service_tags_cli() -> None:
    pass


@service_tags_cli.command(
    name="add",
    help="Add or update tags on a service.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_TAGS_ADD_EXAMPLE,
)
@click.option("--service-id", "--id", help="ID of the service.")
@click.option("--name", "-n", help="Name of the service.")
@click.option("--cloud", type=str, help="Cloud name (for name resolution).")
@click.option("--project", type=str, help="Project name (for name resolution).")
@click.option(
    "--tag",
    "tags",
    multiple=True,
    help="Tag in key=value (or key:value) format. Repeat to add multiple.",
)
def add_tags(
    service_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    tags: Tuple[str],
) -> None:
    if not service_id and not name:
        raise click.ClickException("Provide either --service-id/--id or --name.")
    tag_map = parse_tags_kv_to_str_map(tags)
    if not tag_map:
        raise click.ClickException("Provide at least one --tag key=value.")
    anyscale.service.add_tags(
        id=service_id, name=name, cloud=cloud, project=project, tags=tag_map
    )
    stderr = Console(stderr=True)
    ident = service_id or name or "<unknown>"
    stderr.print(f"Tags updated for service '{ident}'.")


@service_tags_cli.command(
    name="remove",
    help="Remove tags by key from a service.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_TAGS_REMOVE_EXAMPLE,
)
@click.option("--service-id", "--id", help="ID of the service.")
@click.option("--name", "-n", help="Name of the service.")
@click.option("--cloud", type=str, help="Cloud name (for name resolution).")
@click.option("--project", type=str, help="Project name (for name resolution).")
@click.option("--key", "keys", multiple=True, help="Tag key to remove. Repeatable.")
def remove_tags(
    service_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    keys: Tuple[str],
) -> None:
    if not service_id and not name:
        raise click.ClickException("Provide either --service-id/--id or --name.")
    key_list = [k for k in keys if k and k.strip()]
    if not key_list:
        raise click.ClickException("Provide at least one --key to remove.")
    anyscale.service.remove_tags(
        id=service_id, name=name, cloud=cloud, project=project, keys=key_list
    )
    stderr = Console(stderr=True)
    ident = service_id or name or "<unknown>"
    stderr.print(f"Removed tag keys {key_list} from service '{ident}'.")


@service_tags_cli.command(
    name="list",
    help="List tags for a service.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_TAGS_LIST_EXAMPLE,
)
@click.option("--service-id", "--id", help="ID of the service.")
@click.option("--name", "-n", help="Name of the service.")
@click.option("--cloud", type=str, help="Cloud name (for name resolution).")
@click.option("--project", type=str, help="Project name (for name resolution).")
@click.option("--json", "json_output", is_flag=True, default=False, help="JSON output.")
def list_tags(
    service_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    json_output: bool,
) -> None:
    if not service_id and not name:
        raise click.ClickException("Provide either --service-id/--id or --name.")
    tag_map = anyscale.service.list_tags(
        id=service_id, name=name, cloud=cloud, project=project
    )
    if json_output:
        Console().print_json(json=json_dumps(tag_map, indent=2))
    else:
        stderr = Console(stderr=True)
        if not tag_map:
            stderr.print("No tags found.")
            return
        pairs = tag_map.items()
        stderr.print(build_kv_table(pairs, title="Tags"))


# TODO(mowen): Add cloud support for this when we refactor to new SDK method
@service_cli.command(name="rollback", help="Roll back a service.")
@click.option(
    "--service-id", "--id", default=None, help="ID of service.",
)
@click.option("-n", "--name", required=False, default=None, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Max amount of excess capacity allocated during the rollback (0-100).",
)
def rollback(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    max_surge_percent: Optional[int],
):
    """Perform a rollback for a service that is currently in a rollout."""
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id, service_name=name, project_id=project_id,
    )
    service_controller.rollback(service_id, max_surge_percent)


@service_cli.command(name="terminate", help="Terminate a service.")
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option("-n", "--name", required=False, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "-f",
    "--config-file",
    "--service-config-file",
    help="Path to a YAML config file to read the name from. `--service-config-file` is deprecated, use `-f` or `--config-file`.",
)
def terminate(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    config_file: Optional[str],
):
    """Terminate a service.

    This applies to both v1 and v2 services.
    """
    # TODO: Remove service_controller and use the sdk method. Need to update the sdk method
    # so that it can resolve either service name or config file to the service id.
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=config_file,
        project_id=project_id,
    )
    try:
        anyscale.service.terminate(id=service_id)
        log.info(f"Service {service_id} terminate initiated.")
        log.info(
            f"View the service in the UI at {get_endpoint(f'/services/{service_id}')}"
        )
    except Exception as e:  # noqa: BLE001
        log.error(f"Error terminating service: {e}")


@service_cli.command(
    name="archive",
    help="Archive a service.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_ARCHIVE_EXAMPLE,
)
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def archive(
    service_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Archive a service."""

    if (service_id and name) or (not service_id and not name):
        log.error("Exactly one of '--id/--service-id' or '--name' must be provided.")
        sys.exit(1)

    identifier = name if name else service_id
    try:
        anyscale.service.archive(id=service_id, name=name, cloud=cloud, project=project)
        log.info(f"Successfully archived service: {identifier}")
    except Exception as e:  # noqa: BLE001
        log.error(f"Error archiving service: {e}")


@service_cli.command(
    name="delete",
    help="Delete a service.",
    cls=AnyscaleCommand,
    example=command_examples.SERVICE_DELETE_EXAMPLE,
)
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option(
    "-n", "--name", required=False, default=None, type=str, help="Name of the service.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud of this workload. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
def delete(
    service_id: Optional[str],
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Delete a service"""

    if (service_id and name) or (not service_id and not name):
        log.error("Exactly one of '--id/--service-id' or '--name' must be provided.")
        sys.exit(1)

    identifier = name if name else service_id
    try:
        anyscale.service.delete(id=service_id, name=name, cloud=cloud, project=project)
        log.info(f"Successfully deleted service: {identifier}")
    except Exception as e:  # noqa: BLE001
        log.error(f"Error deleting service: {e}")



from typing import List, Optional, Union

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.client.openapi_client.models import (
    AdminCreateUser as AdminCreateUserModel,
    CollaboratorType,
    OrganizationCollaborator,
    OrganizationcollaboratorListResponse,
)
from anyscale.user.models import AdminCreatedUser, AdminCreateUser, User


MAX_PAGE_SIZE = 50


class PrivateUserSDK(BaseSDK):
    @staticmethod
    def _normalize_collaborator_type(
        collaborator_type: Optional[Union[str, CollaboratorType]]
    ) -> Optional[str]:
        if collaborator_type is None:
            return None
        if hasattr(collaborator_type, "value"):
            return collaborator_type.value  # type: ignore[attr-defined]
        return str(collaborator_type)

    @staticmethod
    def _collaborator_to_user(collaborator: OrganizationCollaborator) -> User:
        permission_level = collaborator.permission_level
        if hasattr(permission_level, "value"):
            permission_level = permission_level.value  # type: ignore[attr-defined]

        return User(
            name=collaborator.name,
            email=collaborator.email,
            created_at=collaborator.created_at,
            permission_level=str(permission_level),
            user_id=collaborator.user_id,
        )

    def admin_batch_create(
        self, admin_create_users: List[AdminCreateUser]
    ) -> List[AdminCreatedUser]:
        created_users = self.client.admin_batch_create_users(
            [
                AdminCreateUserModel(**admin_create_user.to_dict())
                for admin_create_user in admin_create_users
            ]
        )

        return [
            AdminCreatedUser(
                user_id=created_user.user_id,
                name=created_user.name,
                email=created_user.email,
                created_at=created_user.created_at,
                is_sso_user=created_user.is_sso_user,
                lastname=created_user.lastname,
                title=created_user.title,
            )
            for created_user in created_users
        ]

    def list(
        self,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        collaborator_type: Optional[Union[str, CollaboratorType]] = None,
        is_service_account: Optional[bool] = None,
        max_items: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> ResultIterator[User]:
        if max_items is not None and max_items <= 0:
            raise ValueError("'max_items' must be greater than 0.")

        if page_size is not None and (page_size <= 0 or page_size > MAX_PAGE_SIZE):
            raise ValueError(
                f"'page_size' must be between 1 and {MAX_PAGE_SIZE}, inclusive."
            )

        def _fetch_page(token: Optional[str],) -> OrganizationcollaboratorListResponse:
            return self.client.list_organization_collaborators(
                email=email,
                name=name,
                collaborator_type=self._normalize_collaborator_type(collaborator_type),
                is_service_account=is_service_account,
                count=page_size,
                paging_token=token,
            )

        return ResultIterator(
            page_token=None,
            max_items=max_items,
            fetch_page=_fetch_page,
            parse_fn=self._collaborator_to_user,
        )

    def get(
        self,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        collaborator_type: Optional[Union[str, CollaboratorType]] = None,
        is_service_account: Optional[bool] = None,
    ) -> User:
        if email is None and name is None:
            raise ValueError("Must provide 'email' or 'name'.")

        response = self.client.list_organization_collaborators(
            email=email,
            name=name,
            collaborator_type=self._normalize_collaborator_type(collaborator_type),
            is_service_account=is_service_account,
            count=1,
        )

        if not response.results:
            target = email if email is not None else name
            raise ValueError(f"No user found matching '{target}'.")

        return self._collaborator_to_user(response.results[0])



from typing import Optional

from anyscale._private.anyscale_client.common import AnyscaleClientInterface
from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk.base_sdk import BaseSDK, Timer
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.create_schedule import CreateSchedule
from anyscale.client.openapi_client.models.decorated_schedule import DecoratedSchedule
from anyscale.client.openapi_client.models.list_response_metadata import (
    ListResponseMetadata,
)
from anyscale.client.openapi_client.models.production_job_config import (
    ProductionJobConfig,
)
from anyscale.client.openapi_client.models.schedule_config import (
    ScheduleConfig as BackendScheduleConfig,
)
from anyscale.job._private.job_sdk import PrivateJobSDK
from anyscale.job.models import JobConfig
from anyscale.schedule.models import ScheduleConfig, ScheduleState, ScheduleStatus


logger = BlockLogger()

MAX_PAGE_SIZE = 50


class PrivateScheduleSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        super().__init__(logger=logger, client=client, timer=timer)
        self._job_sdk = PrivateJobSDK(logger=self.logger, client=self.client)

    def apply(self, config: ScheduleConfig) -> str:
        job_config = config.job_config
        assert isinstance(job_config, JobConfig)
        name = job_config.name or self._job_sdk.get_default_name()

        compute_config_id, cloud_id = self._job_sdk.resolve_compute_config_and_cloud_id(
            compute_config=job_config.compute_config, cloud=job_config.cloud
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=job_config.project
        )

        job_queue_config = None
        if job_config.job_queue_config is not None:
            job_queue_config = self._job_sdk.create_job_queue_config(
                job_config.job_queue_config
            )

        schedule: DecoratedSchedule = self.client.apply_schedule(
            CreateSchedule(
                name=name,
                project_id=project_id,
                config=self._job_sdk.job_config_to_internal_prod_job_conf(
                    config=job_config,
                    name=name,
                    cloud_id=cloud_id,
                    compute_config_id=compute_config_id,
                ),
                job_queue_config=job_queue_config,
                schedule=BackendScheduleConfig(
                    cron_expression=config.cron_expression, timezone=config.timezone,
                ),
            )
        )

        self.logger.info(f"Schedule '{name}' submitted, ID: '{schedule.id}'.")

        return schedule.id

    def _resolve_to_schedule_model(
        self,
        *,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
    ) -> DecoratedSchedule:
        if name is None and id is None:
            raise ValueError("One of 'name' or 'id' must be provided.")

        if name is not None and id is not None:
            raise ValueError("Only one of 'name' or 'id' can be provided.")

        if id is not None and (cloud is not None or project is not None):
            raise ValueError("'cloud' and 'project' should only be used with 'name'.")

        model: Optional[DecoratedSchedule] = self.client.get_schedule(
            name=name, id=id, cloud=cloud, project=project,
        )

        if model is None:
            if name is not None:
                raise RuntimeError(f"Schedule with name '{name}' was not found.")
            else:
                raise RuntimeError(f"Schedule with ID '{id}' was not found.")

        return model

    def set_state(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: ScheduleState,
    ) -> str:
        schedule_model = self._resolve_to_schedule_model(
            name=name, id=id, cloud=cloud, project=project
        )
        is_paused = state == ScheduleState.DISABLED
        self.client.set_schedule_state(id=schedule_model.id, is_paused=is_paused)
        self.logger.info(f"Set schedule '{schedule_model.name}' to state {state}")
        return schedule_model.id

    def _schedule_model_to_status(self, model: DecoratedSchedule) -> ScheduleStatus:
        # TODO: Consider batching get_project calls once the backend API supports
        # fetching multiple projects in a single request.
        project_model = self.client.get_project(model.project_id)
        project = (
            project_model.name
            if project_model is not None and project_model.name != "default"
            else None
        )

        prod_job_config: ProductionJobConfig = model.config
        job_config = self._job_sdk.prod_job_config_to_job_config(
            prod_job_config=prod_job_config, name=model.name, project=project
        )

        config = ScheduleConfig(
            job_config=job_config,
            cron_expression=model.schedule.cron_expression,
            timezone=model.schedule.timezone,
        )

        state = (
            ScheduleState.ENABLED
            if model.next_trigger_at is not None
            else ScheduleState.DISABLED
        )

        return ScheduleStatus(id=model.id, name=model.name, config=config, state=state)

    def status(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> ScheduleStatus:
        schedule_model = self._resolve_to_schedule_model(
            name=name, id=id, cloud=cloud, project=project
        )
        return self._schedule_model_to_status(model=schedule_model)

    def trigger(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        schedule_model = self._resolve_to_schedule_model(
            name=name, id=id, cloud=cloud, project=project
        )
        self.client.trigger_schedule(id=schedule_model.id)
        self.logger.info(f"Triggered job for schedule '{schedule_model.name}'.")
        return schedule_model.id

    def list(
        self,
        *,
        name: Optional[str] = None,
        schedule_id: Optional[str] = None,
        project: Optional[str] = None,
        cloud: Optional[str] = None,
        creator_id: Optional[str] = None,
        page_size: Optional[int] = None,
        max_items: Optional[int] = None,
    ) -> ResultIterator[ScheduleStatus]:
        """List schedules with filtering and pagination.

        Returns a ResultIterator that lazily fetches pages of schedules.
        """
        # Validate page_size
        if page_size is not None and (page_size <= 0 or page_size > MAX_PAGE_SIZE):
            raise ValueError(f"page_size must be between 1 and {MAX_PAGE_SIZE}.")

        # If schedule_id provided, fetch single schedule
        if schedule_id is not None:
            schedule_model = self._resolve_to_schedule_model(id=schedule_id)
            status = self._schedule_model_to_status(model=schedule_model)

            def _fetch_single_page(token: Optional[str]):
                class SingleItemResponse:
                    results = [status] if token is None else []
                    metadata = ListResponseMetadata(total=1, next_paging_token=None)

                return SingleItemResponse()

            return ResultIterator(
                page_token=None,
                max_items=1,
                fetch_page=_fetch_single_page,
                parse_fn=lambda x: x,
            )

        # Resolve cloud and project IDs
        cloud_id = self.client.get_cloud_id(cloud_name=cloud) if cloud else None
        project_id = None
        if project:
            project_id = self.client.get_project_id(
                parent_cloud_id=cloud_id, name=project
            )

        def _fetch_page(token: Optional[str]):
            return self.client.list_schedules(
                name=name,
                project_id=project_id,
                cloud_id=cloud_id,
                creator_id=creator_id,
                count=page_size,
                paging_token=token,
            )

        def _parse_schedule(decorated_schedule: DecoratedSchedule) -> ScheduleStatus:
            return self._schedule_model_to_status(model=decorated_schedule)

        return ResultIterator(
            page_token=None,
            max_items=max_items,
            fetch_page=_fetch_page,
            parse_fn=_parse_schedule,
        )



# type: ignore
"""
This file contains internal APIs extracted from opensource ray.
"""
import logging
import os
from pathlib import Path
from typing import Callable, List, Optional
from zipfile import ZipFile

from pathspec import PathSpec


default_logger = logging.getLogger(__name__)


"""
The following functions are extracted from: ray._private.runtime_env.packaging
"""

# If an individual file is beyond this size, print a warning.
FILE_SIZE_WARNING = 100 * 1024 * 1024  # 100 MiB

DISABLE_GITIGNORE_EXCLUSION = (
    os.environ.get("ANYSCALE_DISABLE_GITIGNORE_EXCLUSION", "0") == "1"
)


def _mib_string(num_bytes: float) -> str:
    size_mib = float(num_bytes / 1024 ** 2)
    return f"{size_mib:.2f}MiB"


def _get_gitignore(path: Path) -> Optional[Callable]:
    if DISABLE_GITIGNORE_EXCLUSION:
        return None

    path = path.absolute()
    ignore_file = path / ".gitignore"
    if ignore_file.is_file():
        with ignore_file.open("r") as f:
            pathspec = PathSpec.from_lines("gitwildmatch", f.readlines())

        def match(p: Path):
            path_str = str(p.absolute().relative_to(path))
            return pathspec.match_file(path_str)

        return match
    else:
        return None


def _dir_travel(
    path: Path,
    excludes: List[Callable],
    handler: Callable,
    logger: Optional[logging.Logger] = default_logger,
):
    """Travels the path recursively, calling the handler on each subpath.
    Respects excludes, which will be called to check if this path is skipped.
    """
    e = _get_gitignore(path)

    if e is not None:
        excludes.append(e)
    skip = any(e(path) for e in excludes)
    if not skip:
        try:
            handler(path)
        except Exception as e:
            logger.error(f"Issue with path: {path}")
            raise e
        if path.is_dir():
            for sub_path in path.iterdir():
                _dir_travel(sub_path, excludes, handler, logger=logger)

    if e is not None:
        excludes.pop()


def _get_excludes(path: Path, excludes: List[str]) -> Callable:
    path = path.absolute()
    pathspec = PathSpec.from_lines("gitwildmatch", excludes)

    def match(p: Path):
        path_str = str(p.absolute().relative_to(path))
        return pathspec.match_file(path_str)

    return match


def zip_directory(
    directory: str,
    excludes: List[str],
    output_path: str,
    include_parent_dir: bool = False,
    logger: Optional[logging.Logger] = default_logger,
) -> None:
    """Zip the target directory and write it to the output_path.
    directory (str): The directory to zip.
    excludes (List(str)): The directories or file to be excluded.
    output_path (str): The output path for the zip file.
    include_parent_dir: If true, includes the top-level directory as a
        directory inside the zip file.
    """
    pkg_file = Path(output_path).absolute()
    with ZipFile(pkg_file, "w") as zip_handler:
        # Put all files in the directory into the zip file.
        dir_path = Path(directory).absolute()

        def handler(path: Path):
            # Pack this path if it's an empty directory or it's a file.
            if (path.is_dir() and next(path.iterdir(), None) is None) or path.is_file():
                file_size = path.stat().st_size
                if file_size >= FILE_SIZE_WARNING:
                    logger.warning(
                        f"File {path} is very large "
                        f"({_mib_string(file_size)}). Consider adding this "
                        "file to the 'excludes' list to skip uploading it: "
                        f"runtime_env={{'excludes': ['{path}', ...], ...}}`"
                    )
                to_path = path.relative_to(dir_path)
                if include_parent_dir:
                    to_path = dir_path.name / to_path
                zip_handler.write(path, to_path)

        excludes = [_get_excludes(dir_path, excludes)]
        _dir_travel(dir_path, excludes, handler, logger=logger)



from typing import Dict, List, Optional

from anyscale._private.anyscale_client import AnyscaleClient
from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk import sdk_command as sdk_command, sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.job_queue._private.job_queue_sdk import PrivateJobQueueSDK
from anyscale.job_queue.commands import (
    _JOB_QUEUE_SDK_SINGLETON_KEY as _JOB_QUEUE_SDK_SINGLETON_KEY,
    _LIST_ARG_DOCSTRINGS,
    _LIST_EXAMPLE,
    _STATUS_ARG_DOCSTRINGS,
    _STATUS_EXAMPLE,
    _TAGS_ADD_ARG_DOCSTRINGS,
    _TAGS_ADD_EXAMPLE,
    _TAGS_LIST_ARG_DOCSTRINGS,
    _TAGS_LIST_EXAMPLE,
    _TAGS_REMOVE_ARG_DOCSTRINGS,
    _TAGS_REMOVE_EXAMPLE,
    _UPDATE_ARG_DOCSTRINGS,
    _UPDATE_EXAMPLE,
    add_tags as add_tags,
    list as list,  # noqa: A004 - claude_comment("claude-opus-4-5", "SDK public API re-export")
    list_tags as list_tags,
    remove_tags as remove_tags,
    status as status,
    update as update,
)
from anyscale.job_queue.models import (
    JobQueueSortDirective,
    JobQueueSortField as JobQueueSortField,
    JobQueueState as JobQueueState,
    JobQueueStatus,
    SessionState,
)


class JobQueueSDK:
    """Public SDK for interacting with Anyscale Job Queues."""

    def __init__(
        self,
        *,
        client: Optional[AnyscaleClient] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateJobQueueSDK(
            client=client, logger=logger, timer=timer
        )

    @sdk_docs(doc_py_example=_LIST_EXAMPLE, arg_docstrings=_LIST_ARG_DOCSTRINGS)
    def list(  # noqa: F811
        self,
        *,
        job_queue_id: Optional[str] = None,
        name: Optional[str] = None,
        creator_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        cluster_status: Optional[SessionState] = None,
        tags_filter: Optional[Dict[str, List[str]]] = None,
        page_size: Optional[int] = None,
        max_items: Optional[int] = None,
        sorting_directives: Optional[List[JobQueueSortDirective]] = None,
    ) -> ResultIterator[JobQueueStatus]:
        """List job queues or fetch a single job queue by ID."""
        return self._private_sdk.list(
            job_queue_id=job_queue_id,
            name=name,
            creator_id=creator_id,
            cloud=cloud,
            project=project,
            cluster_status=cluster_status,
            tags_filter=tags_filter,
            page_size=page_size,
            max_items=max_items,
            sorting_directives=sorting_directives,
        )

    @sdk_docs(doc_py_example=_STATUS_EXAMPLE, arg_docstrings=_STATUS_ARG_DOCSTRINGS)
    def status(self, job_queue_id: str) -> JobQueueStatus:  # noqa: F811
        """Get the status and details for a specific job queue."""
        return self._private_sdk.status(job_queue_id=job_queue_id)

    @sdk_docs(doc_py_example=_UPDATE_EXAMPLE, arg_docstrings=_UPDATE_ARG_DOCSTRINGS)
    def update(  # noqa: F811
        self,
        *,
        job_queue_id: Optional[str] = None,
        job_queue_name: Optional[str] = None,
        max_concurrency: Optional[int] = None,
        idle_timeout_s: Optional[int] = None,
    ) -> JobQueueStatus:
        """Update a job queue."""
        return self._private_sdk.update(
            job_queue_id=job_queue_id,
            job_queue_name=job_queue_name,
            max_concurrency=max_concurrency,
            idle_timeout_s=idle_timeout_s,
        )

    @sdk_docs(doc_py_example=_TAGS_ADD_EXAMPLE, arg_docstrings=_TAGS_ADD_ARG_DOCSTRINGS)
    def add_tags(  # noqa: F811
        self,
        *,
        job_queue_id: Optional[str] = None,
        name: Optional[str] = None,
        tags: Dict[str, str],
    ) -> None:
        """Upsert (add/update) tag key/value pairs for a job queue."""
        return self._private_sdk.add_tags(
            job_queue_id=job_queue_id, name=name, tags=tags
        )

    @sdk_docs(
        doc_py_example=_TAGS_REMOVE_EXAMPLE, arg_docstrings=_TAGS_REMOVE_ARG_DOCSTRINGS
    )
    def remove_tags(  # noqa: F811
        self,
        *,
        job_queue_id: Optional[str] = None,
        name: Optional[str] = None,
        keys: List[str],
    ) -> None:
        """Remove tags by key from a job queue."""
        return self._private_sdk.remove_tags(
            job_queue_id=job_queue_id, name=name, keys=keys
        )

    @sdk_docs(
        doc_py_example=_TAGS_LIST_EXAMPLE, arg_docstrings=_TAGS_LIST_ARG_DOCSTRINGS
    )
    def list_tags(  # noqa: F811
        self, *, job_queue_id: Optional[str] = None, name: Optional[str] = None,
    ) -> Dict[str, str]:
        """List tags for a job queue."""
        return self._private_sdk.list_tags(job_queue_id=job_queue_id, name=name)



from typing import Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.cloud_utils import get_cloud_id_and_name
from anyscale.controllers.base_controller import BaseController


class ExperimentalIntegrationsController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")

    # TODO(nikita): Use discriminated union types for cloud_id/cloud_name
    def enable_wandb_integration(
        self, cloud_id: Optional[str], cloud_name: Optional[str]
    ):
        """
        [DEPRECATED]
        Enables W&B integration for the current user for the provided cloud. This
        is currently only implemented for AWS clouds.
        """
        feature_flag_on = self.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
            "wandb-integration-prototype"
        ).result.is_on
        if not feature_flag_on:
            raise click.ClickException(
                "The W&B integration can only be enabled if the feature flag is enabled. "
                "Please contact Anyscale support to enable the flag."
            )

        # Assumes only one of cloud_id and cloud_name is passed. This should be checked with
        # a user friendly error message at the command layer.
        assert (
            bool(cloud_id) + bool(cloud_name) == 1
        ), "Must provide only one of --cloud_id or --cloud_name."

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )
        cloud = self.api_client.get_cloud_api_v2_clouds_cloud_id_get(
            cloud_id=cloud_id
        ).result

        user_id = self.api_client.get_user_info_api_v2_userinfo_get().result.id

        secret_name = f"wandb_api_key_{user_id}"
        cloud_region = cloud.region
        if cloud.provider == "AWS":
            put_secret_cmd = f"`AWS_DEFAULT_REGION={cloud_region} aws secretsmanager create-secret --name {secret_name} --secret-string {{my-wandb-api-key}}`"
        elif cloud.provider == "GCP":
            put_secret_cmd = f'printf "{{my-wandb-api-key}}" | gcloud secrets create {secret_name} --data-file=-'
        else:
            raise click.ClickException(
                "W&B integration is currently only supported for AWS and GCP clouds."
            )
        steps = f"""
            1. Place your W&B API key in the AWS secrets manager associated with your Anyscale cloud account. The secret key should be `{secret_name}`.
               This can be done by running {self.log.highlight(put_secret_cmd)}
               in an environment that can write secrets to your {cloud.provider} account.
            2. [Optional] Allow Anyscale clusters started in {cloud_name} to access your secret store for the cloud by running {self.log.highlight(f"`anyscale cloud secrets --name {cloud_name}`")}.
               This is not necessary if the organization admin already granted access.
            3. Run a production job or workspace using ray.air.callbacks.wandb.WandbLoggerCallback in your code. No parameters need to be passed to WandbLoggerCallback to use this integration.
               \033[1mNote:\033[0m Please make sure to restart any existing workspace after running this command to ensure the integration is enabled.
            """

        self.log.info(
            f"Enabled W&B integration for cloud {cloud_name}. Please follow the steps below to use the integration:\n{steps}"
        )

        # TODO(nikita): Link to documentation here
        # docs_url = "N/A"
        # self.log.info(f"Please find more detailed instructions on using this integration at {docs_url}")



from abc import ABC, abstractmethod
import contextlib
from datetime import datetime
from functools import wraps
import io
import json
import logging
import os
import pathlib
import re
import time
from typing import Any, Callable, Dict, Generator, IO, List, Optional, Tuple
from urllib.parse import urlparse

from openapi_client.exceptions import ApiException
import requests
from rich.style import Style
import smart_open

from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface,
    DEFAULT_PYTHON_VERSION,
    DEFAULT_RAY_VERSION,
    RUNTIME_ENV_PACKAGE_FORMAT,
)
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.utils.progress_util import FileDownloadProgress
from anyscale.api_utils.logs_util import _download_log_from_s3_url_sync
from anyscale.authenticate import AuthenticationBlock, get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models import (
    AdminCreatedUser,
    AdminCreateUser,
    AnyscaleServiceAccount,
    ApiKeyParameters,
    ApplyProductionServiceMultiVersionV2Model,
    ArchiveStatus,
    Cloud,
    CloudDataBucketAccessMode,
    CloudDataBucketFileType,
    CloudDataBucketPresignedUrlRequest,
    CloudDataBucketPresignedUrlResponse,
    CloudDataBucketPresignedUrlScheme,
    CloudNameOptions,
    ClusteroperationResponse,
    CollaboratorType,
    ComputeTemplate,
    ComputeTemplateConfig,
    ComputeTemplateQuery,
    CreateCloudCollaborator,
    CreateComputeTemplate,
    CreateExperimentalWorkspace,
    CreateInternalProductionJob,
    CreateOrganizationInvitation,
    CreateResourceQuota,
    CreateUserProjectCollaborator,
    DecoratedCloudResource,
    DecoratedComputeTemplate,
    DecoratedjobqueueListResponse,
    DecoratedlistserviceapimodelListResponse,
    DecoratedproductionjobListResponse,
    DecoratedProductionServiceV2APIModel,
    DecoratedProductionServiceV2VersionAPIModel,
    DecoratedSession,
    DeleteResourceTagsRequest,
    ExperimentalWorkspace,
    GetOrCreateBuildFromImageUriRequest,
    InternalProductionJob,
    JobQueueSortDirective,
    JobQueuesQuery,
    ListResourceQuotasQuery,
    OrganizationCollaborator,
    OrganizationcollaboratorListResponse,
    OrganizationInvitation,
    PolicyResponse,
    Project,
    ProjectBase,
    ProjectListResponse,
    ResourcepolicyitemListResponse,
    ResourceQuota,
    ResourceQuotaStatus,
    ResourceTagResourceType,
    ServerSessionToken,
    SessionSshKey,
    SessionState,
    StartSessionOptions,
    StopSessionOptions,
    SystemWorkloadName,
    UpdatePolicyRequest,
    UpsertResourceTagsRequest,
    UserGroup,
    UsergroupListResponse,
    UserInfo,
    WorkspaceDataplaneProxiedArtifacts,
    WriteProject,
)
from anyscale.client.openapi_client.models.create_schedule import CreateSchedule
from anyscale.client.openapi_client.models.decorated_application_template import (
    DecoratedApplicationTemplate,
)
from anyscale.client.openapi_client.models.decorated_job_queue import DecoratedJobQueue
from anyscale.client.openapi_client.models.decorated_schedule import DecoratedSchedule
from anyscale.client.openapi_client.models.decoratedapplicationtemplate_list_response import (
    DecoratedapplicationtemplateListResponse,
)
from anyscale.client.openapi_client.models.decoratedschedule_list_response import (
    DecoratedscheduleListResponse,
)
from anyscale.client.openapi_client.models.production_job import ProductionJob
from anyscale.client.openapi_client.models.resource_tag_record import ResourceTagRecord
from anyscale.client.openapi_client.models.update_job_queue_request import (
    UpdateJobQueueRequest,
)
from anyscale.client.openapi_client.rest import ApiException as InternalApiException
from anyscale.cluster_compute import parse_cluster_compute_name_version
from anyscale.feature_flags import FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as ExternalApi
from anyscale.sdk.anyscale_client.models import (
    ApplyProductionServiceV2Model,
    Cluster,
    ClusterCompute,
    ClusterComputeConfig,
    ClusterComputesQuery,
    ClusterEnvironment,
    ClusterEnvironmentBuild,
    ClusterenvironmentbuildListResponse,
    ClusterEnvironmentBuildStatus,
    ClusterEnvironmentsQuery,
    CreateBYODClusterEnvironment,
    CreateBYODClusterEnvironmentConfigurationSchema,
    CreateClusterEnvironment,
    CreateClusterEnvironmentBuild,
    Job as APIJobRun,
    ProductionServiceV2VersionModel,
    Project as ProjectExternal,
    RollbackServiceModel,
    TextQuery,
)
from anyscale.sdk.anyscale_client.models.jobs_query import JobsQuery
from anyscale.sdk.anyscale_client.models.jobs_sort_field import JobsSortField
from anyscale.sdk.anyscale_client.models.page_query import PageQuery
from anyscale.sdk.anyscale_client.models.sort_by_clause_jobs_sort_field import (
    SortByClauseJobsSortField,
)
from anyscale.sdk.anyscale_client.models.sort_order import SortOrder
from anyscale.sdk.anyscale_client.rest import ApiException as ExternalApiException
from anyscale.shared_anyscale_utils.bytes_util import Bytes
from anyscale.shared_anyscale_utils.conf import ANYSCALE_HOST
from anyscale.shared_anyscale_utils.latest_ray_version import LATEST_RAY_VERSION
from anyscale.util import (
    get_cluster_model_for_current_workspace,
    get_endpoint,
    is_anyscale_workspace,
)
from anyscale.utils.connect_helpers import search_entities
from anyscale.utils.runtime_env import (
    is_workspace_dependency_tracking_disabled,
    parse_dot_env_file,
    WORKSPACE_REQUIREMENTS_FILE_PATH,
    zip_local_dir,
)
from anyscale.utils.workspace_notification import (
    WORKSPACE_NOTIFICATION_ADDRESS,
    WorkspaceNotification,
)


WORKSPACE_ID_ENV_VAR = "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID"
OVERWRITE_EXISTING_CLOUD_STORAGE_FILES = (
    os.environ.get("ANYSCALE_OVERWRITE_EXISTING_CLOUD_STORAGE_FILES", "0") == "1"
)

# internal_logger is used for logging internal errors or debug messages that we do not expect users to see.
internal_logger = logging.getLogger(__name__)

# A decorator to handle ApiException and raise ValueError with the error message.
def handle_api_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (ApiException, InternalApiException, ExternalApiException) as e:
            if e.status >= 400 and e.status < 500:
                try:
                    body_dict = json.loads(e.body)
                    msg = body_dict["error"]["detail"]
                    raise ValueError(msg) from None
                except (KeyError, TypeError):
                    # ApiException doesn't conform to expected format, raise original error
                    raise e from None
            raise e from None

    return wrapper


@contextlib.contextmanager
def set_env(**environ):
    """
    Temporarily set the process environment variables.

    >>> with set_env(PLUGINS_DIR='test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True

    >>> "PLUGINS_DIR" in os.environ
    False

    :type environ: dict[str, unicode]
    :param environ: Environment variables to set
    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


class AWSS3ClientInterface(ABC):
    @abstractmethod
    def download_fileobj(self, Bucket: str, Key: str, Fileobj: IO[Any],) -> None:
        """Download a file from an S3 bucket to a file-like object."""
        raise NotImplementedError


class GCSBlobInterface(ABC):
    @abstractmethod
    def download_to_file(self, fileobj: IO[Any]) -> None:
        """Download the blob to a file-like object."""
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        """Check if the blob exists."""
        raise NotImplementedError


class GCSBucketInterface(ABC):
    @abstractmethod
    def blob(self, object_name: str) -> GCSBlobInterface:
        """Get a blob object for the given object name."""
        raise NotImplementedError


class GCPGCSClientInterface(ABC):
    @abstractmethod
    def bucket(self, bucket: str) -> GCSBucketInterface:
        """Get a bucket object for the given bucket name."""
        raise NotImplementedError


class AnyscaleClient(AnyscaleClientInterface):
    # Number of entries to fetch per request for list endpoints.
    LIST_ENDPOINT_COUNT = 50

    def __init__(
        self,
        *,
        api_clients: Optional[Tuple[ExternalApi, InternalApi]] = None,
        sleep: Optional[Callable[[float], None]] = None,
        workspace_requirements_file_path: str = WORKSPACE_REQUIREMENTS_FILE_PATH,
        logger: Optional[BlockLogger] = None,
        host: Optional[str] = None,
        s3_client: Optional[AWSS3ClientInterface] = None,
        gcs_client: Optional[GCPGCSClientInterface] = None,
    ):
        if api_clients is None:
            auth_block: AuthenticationBlock = get_auth_api_client(
                raise_structured_exception=True
            )
            api_clients = (auth_block.anyscale_api_client, auth_block.api_client)

        self._external_api_client, self._internal_api_client = api_clients
        self._workspace_requirements_file_path = workspace_requirements_file_path
        self._sleep = sleep or time.sleep
        self._s3_client = s3_client
        self._gcs_client = gcs_client

        # Cached IDs and models to avoid duplicate lookups.
        self._default_project_id_from_cloud_id: Dict[Optional[str], str] = {}
        self._cloud_id_cache: Dict[Optional[str], str] = {}
        self._cluster_env_build_cache: Dict[str, ClusterEnvironmentBuild] = {}
        self._current_workspace_cluster: Optional[Cluster] = None
        self._logger = logger or BlockLogger()
        self._host = host or ANYSCALE_HOST

    @property
    def s3_client(self) -> AWSS3ClientInterface:
        if self._s3_client is None:
            # initialize the s3 client lazily so that we import the boto3 library only when needed.
            try:
                import boto3  # noqa: PLC0415 - codex_reason("gpt5.2", "optional AWS SDK dependency")
                import botocore.config  # noqa: PLC0415 - codex_reason("gpt5.2", "optional AWS SDK dependency")
            except ImportError:
                raise RuntimeError(
                    "Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
                )
            self._s3_client = boto3.client(  # type: ignore
                "s3", config=botocore.config.Config(signature_version="s3v4")
            )
        return self._s3_client  # type: ignore

    @property
    def gcs_client(self) -> GCPGCSClientInterface:
        if self._gcs_client is None:
            # initialize the gcs client lazily so that we import the google cloud storage library only when needed.
            try:
                from google.cloud import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional GCP storage dependency")
                    storage,
                )
            except ImportError:
                raise RuntimeError(
                    "Could not import the Google Storage Python API via `from google.cloud import storage`.  Please check your installation or try running `pip install --upgrade google-cloud-storage`."
                )
            self._gcs_client = storage.Client()
        return self._gcs_client  # type: ignore

    @property
    def host(self) -> str:
        return self._host

    @property
    def logger(self) -> BlockLogger:
        return self._logger

    def get_job_ui_url(self, job_id: str) -> str:
        return get_endpoint(f"/jobs/{job_id}", host=self.host)

    def get_service_ui_url(self, service_id: str) -> str:
        return get_endpoint(f"/services/{service_id}", host=self.host)

    def get_compute_config_ui_url(
        self, compute_config_id: str, *, cloud_id: str
    ) -> str:
        return get_endpoint(
            f"/v2/{cloud_id}/compute-configs/{compute_config_id}", host=self.host
        )

    def get_build_ui_url(self, cluster_env_id, build_id: str) -> str:
        return get_endpoint(
            f"v2/container-images/{cluster_env_id}/versions/{build_id}", host=self.host
        )

    def get_current_workspace_id(self) -> Optional[str]:
        return os.environ.get(WORKSPACE_ID_ENV_VAR, None)

    def inside_workspace(self) -> bool:
        return self.get_current_workspace_id() is not None

    def get_workspace_requirements_path(self) -> Optional[str]:
        if (
            not self.inside_workspace()
            or is_workspace_dependency_tracking_disabled()
            or not pathlib.Path(self._workspace_requirements_file_path).is_file()
        ):
            return None

        return self._workspace_requirements_file_path

    def _download_file_from_google_cloud_storage(
        self, bucket: str, object_name: str
    ) -> Optional[bytes]:
        try:
            bucket_obj = self.gcs_client.bucket(bucket)
            blob = bucket_obj.blob(object_name)
            fileobj = io.BytesIO()
            if blob.exists():
                blob.download_to_file(fileobj)
                return fileobj.getvalue()
            return None
        except Exception as e:  # noqa: BLE001
            raise RuntimeError(
                f"Failed to download the working directory from Google Cloud Storage. Error {e!r}"
                "Please validate you have exported cloud credentials with the correct read permissions and the intended bucket exists in your Cloud Storage account. "
            ) from e

    def _download_file_from_s3(self, bucket: str, object_key: str) -> Optional[bytes]:
        try:
            from botocore.exceptions import (  # noqa: PLC0415 - codex_reason("gpt5.2", "optional AWS SDK dependency")
                ClientError,
            )
        except Exception:  # noqa: BLE001
            raise RuntimeError(
                "Could not download file from S3: Could not import the Amazon S3 Python API via `import boto3`.  Please check your installation or try running `pip install boto3`."
            )
        try:
            fileobj = io.BytesIO()
            self.s3_client.download_fileobj(bucket, object_key, fileobj)
            return fileobj.getvalue()
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return None
            raise
        except Exception as e:  # noqa: BLE001
            raise RuntimeError(
                f"Failed to download the working directory from S3. Error {e!r}"
                "Please validate you have exported cloud credentials with the correct read permissions and the intended bucket exists in your S3 account. "
            ) from e

    def _download_file_from_remote_storage(self, remote_uri: str) -> Optional[bytes]:
        parsed_uri = urlparse(remote_uri)
        service = parsed_uri.scheme
        bucket = parsed_uri.netloc
        object_name = parsed_uri.path.lstrip("/")
        if service == "s3":
            return self._download_file_from_s3(bucket, object_name)
        if service == "gs":
            return self._download_file_from_google_cloud_storage(bucket, object_name)
        return None

    def get_workspace_env_vars(self) -> Optional[Dict[str, str]]:
        system_storage_path = os.environ.get("ANYSCALE_INTERNAL_SYSTEM_STORAGE", "")
        workspace_id = os.environ.get(WORKSPACE_ID_ENV_VAR, "")
        workspace_artifact_path = (
            os.path.join(
                system_storage_path, "workspace_tracking_dependencies", workspace_id,
            )
            if workspace_id and system_storage_path
            else ""
        )
        workspace_dot_env_path = (
            os.path.join(workspace_artifact_path, ".env")
            if workspace_artifact_path
            else ""
        )

        if not self.inside_workspace() or not workspace_dot_env_path:
            return None

        dot_env = self._download_file_from_remote_storage(workspace_dot_env_path)
        if dot_env:
            parsed_dot_env = parse_dot_env_file(dot_env)
            if parsed_dot_env:
                self.logger.info(
                    f"Using workspace runtime dependencies env vars: {parsed_dot_env}."
                )
            return parsed_dot_env
        return None

    @handle_api_exceptions
    def get_current_workspace_cluster(self) -> Optional[Cluster]:
        # Checks for the existence of the ANYSCALE_EXPERIMENTAL_WORKSPACE_ID env var.
        if not is_anyscale_workspace():
            return None

        if self._current_workspace_cluster is None:
            # Picks up the cluster ID from the ANYSCALE_SESSION_ID env var.
            self._current_workspace_cluster = get_cluster_model_for_current_workspace(
                self._external_api_client
            )

        return self._current_workspace_cluster

    def _get_project_id_by_name(
        self, *, parent_cloud_id: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        """Resolves `name`, a project name, to a project ID.

        Args:
            parent_cloud_id: If specified, return the project that has this `parent_cloud_id`. \
                Else (`None`), return the first project with the given `name`. \
                Defaults to `None`.
            name: The name of the project.
        """
        # First try to find project by name and parent cloud id
        if parent_cloud_id is not None and name is not None:
            response = self._external_api_client.search_projects(
                {
                    "name": {"equals": name,},
                    "parent_cloud_id": {"equals": parent_cloud_id,},
                }
            )
            if response.results:
                # Project name is unique within a cloud, so we can just return the first result
                return response.results[0].id

        # Then find if project with name already exists
        matching_projects = self._internal_api_client.find_project_by_project_name_api_v2_projects_find_by_name_get(
            name
        ).results
        if len(matching_projects) == 0:
            raise ValueError(f"Project '{name}' was not found.")
        else:
            for project in matching_projects:
                if (
                    parent_cloud_id is None
                    or project.parent_cloud_id == parent_cloud_id
                ):
                    return project.id
            raise ValueError(
                f"{len(matching_projects)} project(s) found with name '{name}' and none matched cloud_id '{parent_cloud_id}'"
            )

    def _get_project_id_by_cloud_id(
        self, *, parent_cloud_id: Optional[str] = None,
    ) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            if (
                workspace_cluster.cluster_compute_config is not None
                and workspace_cluster.cluster_compute_config.cloud_id == parent_cloud_id
            ):
                return workspace_cluster.project_id
            elif workspace_cluster.cluster_compute_id is not None:
                workspace_cluster_compute = self.get_compute_config(
                    workspace_cluster.cluster_compute_id
                )
                if (
                    workspace_cluster_compute is not None
                    and workspace_cluster_compute.config is not None
                    and workspace_cluster_compute.config.cloud_id == parent_cloud_id
                ):
                    return workspace_cluster.project_id

        if self._default_project_id_from_cloud_id.get(parent_cloud_id) is None:
            # Cloud isolation organizations follow the permissions model in https://docs.anyscale.com/organization-and-user-account/access-controls
            default_project: ProjectExternal = self._external_api_client.get_default_project(
                parent_cloud_id=parent_cloud_id
            ).result
            self._default_project_id_from_cloud_id[parent_cloud_id] = default_project.id

        return self._default_project_id_from_cloud_id[parent_cloud_id]

    @handle_api_exceptions
    def get_project_id(
        self, *, parent_cloud_id: Optional[str] = None, name: Optional[str] = None
    ) -> str:
        if name is not None:
            return self._get_project_id_by_name(
                parent_cloud_id=parent_cloud_id, name=name
            )
        else:
            return self._get_project_id_by_cloud_id(parent_cloud_id=parent_cloud_id)

    def _get_cloud_id_for_compute_config_id(self, compute_config_id: str) -> str:
        cluster_compute: ClusterCompute = self._external_api_client.get_cluster_compute(
            compute_config_id
        ).result
        cluster_compute_config: ClusterComputeConfig = cluster_compute.config
        return cluster_compute_config.cloud_id

    def _get_cloud_id_by_name(self, cloud_name: str) -> Optional[str]:
        try:
            return self._internal_api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post(
                CloudNameOptions(name=cloud_name),
            ).result.id
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def get_cloud_id(
        self, cloud_name: Optional[str] = None, compute_config_id: Optional[str] = None
    ) -> str:
        if cloud_name is not None and compute_config_id is not None:
            raise ValueError(
                "Only one of cloud_name or compute_config_id should be provided."
            )

        if compute_config_id is not None:
            return self._get_cloud_id_for_compute_config_id(compute_config_id)

        if cloud_name in self._cloud_id_cache:
            return self._cloud_id_cache[cloud_name]

        if cloud_name is not None:
            cloud_id = self._get_cloud_id_by_name(cloud_name)
            if cloud_id is None:
                raise RuntimeError(f"Cloud '{cloud_name}' not found.")
        elif self.inside_workspace():
            workspace_cluster = self.get_current_workspace_cluster()
            assert workspace_cluster is not None
            # NOTE(edoakes): the Cluster model has a compute_config_config model that includes
            # its cloud ID, but it's not always populated.
            # TODO(edoakes): add cloud_id to the Cluster model to avoid a second RTT.
            cloud_id = self._get_cloud_id_for_compute_config_id(
                workspace_cluster.cluster_compute_id
            )
        else:
            cloud_id = self._external_api_client.get_default_cloud().result.id

        assert cloud_id is not None
        self._cloud_id_cache[cloud_name] = cloud_id
        return cloud_id

    @handle_api_exceptions
    def get_cloud(self, *, cloud_id: str) -> Optional[Cloud]:
        try:
            cloud: Cloud = self._external_api_client.get_cloud(cloud_id=cloud_id).result
            return cloud
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def get_cloud_by_name(self, *, name) -> Optional[Cloud]:
        cloud_id = self.get_cloud_id(cloud_name=name)
        return self.get_cloud(cloud_id=cloud_id)

    @handle_api_exceptions
    def get_cloud_resource_by_name(
        self, cloud_id: str, cloud_resource_name: str
    ) -> Optional[DecoratedCloudResource]:
        return self._internal_api_client.find_cloud_resource_by_name_api_v2_clouds_cloud_id_find_cloud_resource_by_name_post(
            cloud_id=cloud_id, cloud_resource_name=cloud_resource_name,
        ).result

    @handle_api_exceptions
    def get_default_cloud(self) -> Optional[Cloud]:
        try:
            return self._external_api_client.get_default_cloud().result
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def list_clouds(
        self, *, paging_token: Optional[str] = None, count: Optional[int] = None
    ):
        return self._internal_api_client.list_clouds_api_v2_clouds_get(
            paging_token=paging_token, count=count or self.LIST_ENDPOINT_COUNT
        )

    @handle_api_exceptions
    def add_cloud_collaborators(
        self, cloud_id: str, collaborators: List[CreateCloudCollaborator]
    ) -> None:
        self._internal_api_client.batch_create_cloud_collaborators_api_v2_clouds_cloud_id_collaborators_users_batch_create_post(
            cloud_id, collaborators
        )

    @handle_api_exceptions
    def terminate_system_cluster(self, cloud_id: str) -> ClusteroperationResponse:
        return self._internal_api_client.terminate_system_cluster_api_v2_system_workload_cloud_id_terminate_post(
            cloud_id
        )

    @handle_api_exceptions
    def describe_system_workload_get_status(self, cloud_id: str) -> str:
        res = self._internal_api_client.describe_system_workload_api_v2_system_workload_cloud_id_describe_post(
            cloud_id, SystemWorkloadName.RAY_OBS_EVENTS_API_SERVICE, start_cluster=False
        ).result
        return res.status

    @handle_api_exceptions
    def create_compute_config(
        self, config: ComputeTemplateConfig, *, name: Optional[str] = None
    ) -> Tuple[str, str]:
        result: ComputeTemplate = self._internal_api_client.create_compute_template_api_v2_compute_templates_post(
            create_compute_template=CreateComputeTemplate(
                config=config, name=name, anonymous=name is None, new_version=True
            )
        ).result
        return f"{result.name}:{result.version}", result.id

    @handle_api_exceptions
    def get_compute_config(
        self, compute_config_id: str
    ) -> Optional[DecoratedComputeTemplate]:
        try:
            return self._internal_api_client.get_compute_template_api_v2_compute_templates_template_id_get(
                compute_config_id
            ).result
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def get_compute_config_id(
        self,
        compute_config_name: Optional[str] = None,
        cloud: Optional[str] = None,
        *,
        include_archived: bool = False,
    ) -> Optional[str]:
        if compute_config_name is not None:
            name, version = parse_cluster_compute_name_version(compute_config_name)
            if version is None:
                # Setting `version=-1` will return only the latest version if there are multiple.
                version = -1

            cloud_id = (
                self.get_cloud_id(cloud_name=cloud) if cloud is not None else None
            )

            cluster_computes = self._internal_api_client.search_compute_templates_api_v2_compute_templates_search_post(
                ComputeTemplateQuery(
                    orgwide=True,
                    name={"equals": name},
                    include_anonymous=True,
                    archive_status=ArchiveStatus.ALL
                    if include_archived
                    else ArchiveStatus.NOT_ARCHIVED,
                    version=version,
                    cloud_id=cloud_id,
                )
            ).results

            if len(cluster_computes) == 0:
                return None

            compute_template: DecoratedComputeTemplate = cluster_computes[0]
            return compute_template.id

        # If the compute config name is not provided, we pick an appropriate default.
        #
        #   - If running in a workspace:
        #       * If auto_select_worker_config enabled: we switch over to a standardized
        #         default compute config (copying over any cluster-level attributes, e.g.
        #         max-gpus).
        #       * Otherwise, we use the workspace's compute config.
        #
        #   - Otherwise, we use the default compute config provided by the API.

        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            workspace_compute_config: DecoratedComputeTemplate = self.get_compute_config(
                workspace_cluster.cluster_compute_id
            )
            workspace_config: ClusterComputeConfig = workspace_compute_config.config
            if workspace_config.auto_select_worker_config:
                standard_template = self._build_standard_compute_template_from_existing_auto_config(
                    workspace_config
                )
                _, compute_config_id = self.create_compute_config(standard_template)
                return compute_config_id
            else:
                return workspace_cluster.cluster_compute_id

        return self.get_default_compute_config(cloud_id=self.get_cloud_id()).id

    @handle_api_exceptions
    def archive_compute_config(self, *, compute_config_id):
        self._internal_api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post(
            compute_config_id
        )

    @handle_api_exceptions
    def search_cluster_computes(self, query: Dict[str, Any]):
        """Search for cluster computes matching the provided query.

        Uses the external API client (SDK) to search for cluster computes.
        """

        # Convert dict to ClusterComputesQuery model
        cluster_computes_query = ClusterComputesQuery(**query)
        return self._external_api_client.search_cluster_computes(cluster_computes_query)

    @handle_api_exceptions
    def get_default_compute_config(
        self, *, cloud_id: str, cloud_resource_id: Optional[str] = None
    ) -> ClusterCompute:
        return self._external_api_client.get_default_cluster_compute(
            cloud_id=cloud_id, cloud_resource_id=cloud_resource_id
        ).result

    def _build_standard_compute_template_from_existing_auto_config(
        self, compute_config: ClusterComputeConfig
    ) -> ComputeTemplateConfig:
        """
        Build a standard compute template config from an existing compute config.

        1. Pull the default compute template config.
        2. Disable scheduling on the head node.
        3. Enable auto_select_worker_config.
        4. Copy any cluster-level flags from the provided compute config to the template.
        """
        # Retrieve the default cluster compute config for the cloud.
        default_compute_template: DecoratedComputeTemplate = self._external_api_client.get_default_cluster_compute(
            cloud_id=compute_config.cloud_id,
        ).result.config

        # Disable head node scheduling.
        if default_compute_template.head_node_type.resources is None:
            default_compute_template.head_node_type.resources = {}
        default_compute_template.head_node_type.resources["CPU"] = 0

        # Ensure auto_select_worker_config is enabled.
        default_compute_template.auto_select_worker_config = True

        # Copy flags set at a cluster level over to the workspace.
        #
        # NOTE (shomilj): If there are more attributes we want to
        # persist from the provided compute config --> the compute
        # config used for deploying the service, we should copy them
        # over here.
        default_compute_template.flags = compute_config.flags

        return default_compute_template

    @handle_api_exceptions
    def get_cluster_env_build(self, build_id: str) -> Optional[ClusterEnvironmentBuild]:
        # Check cache first
        if build_id in self._cluster_env_build_cache:
            return self._cluster_env_build_cache[build_id]

        # Fetch from API if not in cache
        try:
            res = self._external_api_client.get_cluster_environment_build(build_id)
            build = res.result
        except ExternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

        # Store in cache ONLY if the build exists and is in a terminal state
        if build:
            terminal_states = {
                ClusterEnvironmentBuildStatus.SUCCEEDED,
                ClusterEnvironmentBuildStatus.FAILED,
                ClusterEnvironmentBuildStatus.CANCELED,
            }
            if build.status in terminal_states:
                self._cluster_env_build_cache[build_id] = build

        return build

    @handle_api_exceptions
    def get_cluster_env_build_image_uri(
        self, cluster_env_build_id: str, use_image_alias: bool = False
    ) -> Optional[ImageURI]:
        try:
            build = self.get_cluster_env_build(cluster_env_build_id)
            if build is None:
                return None

            cluster_env = self._external_api_client.get_cluster_environment(
                build.cluster_environment_id
            ).result
            return ImageURI.from_cluster_env_build(cluster_env, build, use_image_alias)
        except ExternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def archive_image(self, *, image_id: str) -> None:
        """Archive an image (cluster environment) by ID."""
        self._internal_api_client.archive_cluster_environment_api_v2_application_templates_application_template_id_archive_post(
            image_id
        )

    @handle_api_exceptions
    def get_default_build_id(self) -> str:
        workspace_cluster = self.get_current_workspace_cluster()
        if workspace_cluster is not None:
            return workspace_cluster.cluster_environment_build_id
        result: ClusterEnvironmentBuild = self._external_api_client.get_default_cluster_environment_build(
            DEFAULT_PYTHON_VERSION, DEFAULT_RAY_VERSION,
        ).result
        return result.id

    @handle_api_exceptions
    def get_cluster_env_by_name(self, name: str) -> Optional[ClusterEnvironment]:
        resp = self._external_api_client.search_cluster_environments(
            ClusterEnvironmentsQuery(
                name=TextQuery(equals=name),  # pyright: ignore reportGeneralTypeIssues
                paging=PageQuery(count=1),
                include_anonymous=True,
            )
        )
        if resp.results:
            return resp.results[0]
        return None

    @handle_api_exceptions
    def get_application_template(
        self, application_template_id: str
    ) -> Optional[DecoratedApplicationTemplate]:
        try:
            return self._internal_api_client.get_application_template_api_v2_application_templates_application_template_id_get(
                application_template_id
            ).result
        except InternalApiException as e:
            if e.status == 404:
                return None
            raise e from None

    @handle_api_exceptions
    def list_application_templates(  # noqa: PLR0913
        self,
        *,
        name: Optional[str] = None,
        image_name: Optional[str] = None,
        creator_id: Optional[str] = None,
        project: Optional[str] = None,
        include_archived: bool = False,
        defaults_first: bool = False,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
    ) -> DecoratedapplicationtemplateListResponse:
        project_id = None
        if project:
            project_id = self.get_project_id(name=project)

        return self._internal_api_client.list_application_templates_api_v2_application_templates_get(
            project_id=project_id,
            creator_id=creator_id,
            name_contains=name,
            image_name_contains=image_name,
            include_archived=include_archived,
            defaults_first=defaults_first,
            paging_token=paging_token,
            count=count if count is not None else self.LIST_ENDPOINT_COUNT,
        )

    @handle_api_exceptions
    def list_cluster_env_builds(
        self, cluster_env_id: str,
    ) -> Generator[ClusterEnvironmentBuild, None, None]:
        paging_token = None
        while True:
            resp: ClusterenvironmentbuildListResponse = self._external_api_client.list_cluster_environment_builds(
                cluster_environment_id=cluster_env_id,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
                desc=True,
            )
            for build in resp.results:
                yield build

            if resp.metadata.next_paging_token is None:
                break
            paging_token = resp.metadata.next_paging_token

    def _wait_for_build_to_succeed(
        self, build_id: str, poll_interval_seconds=3, timeout_secs=3600,
    ):
        """Periodically check the status of the build operation until it completes.
        Raise a RuntimeError if the build fails or cancelled.
        Raise a TimeoutError if the build does not complete within the timeout.
        """
        elapsed_secs = 0
        while elapsed_secs < timeout_secs:
            build = self.get_cluster_env_build(build_id)
            if build.status == ClusterEnvironmentBuildStatus.SUCCEEDED:
                self.logger.info("")
                return
            elif build.status == ClusterEnvironmentBuildStatus.FAILED:
                raise RuntimeError(f"Image build {build_id} failed.")
            elif build.status == ClusterEnvironmentBuildStatus.CANCELED:
                raise RuntimeError(f"Image build {build_id} unexpectedly cancelled.")

            elapsed_secs += poll_interval_seconds
            self.logger.info(
                f"Waiting for image build to complete. Elapsed time: {elapsed_secs} seconds.",
                end="\r",
            )
            self._sleep(poll_interval_seconds)
        raise TimeoutError(
            f"Timed out waiting for image build {build_id} to complete after {timeout_secs}s."
        )

    def _find_or_create_cluster_env(
        self,
        cluster_env_name: str,
        anonymous: bool,
        *,
        image_uri: Optional[str] = None,
        registry_login_secret: Optional[str] = None,
        ray_version: Optional[str] = None,
    ) -> ClusterEnvironment:
        """
        Find or create a cluster environment with the given name.

        There're two possible race conditions:
        1) A tries to create a cluster env with the same name as B, but B has already created it.
        -> A will get a 409 conflict error and should retry to get the existing cluster env.
        2) A and B creates two identical builds under the same cluster env. This would cause job queue to
        reject the job submission.
        -> Cluster env and BYOD build are created within the same transaction, so the latter one will fail
        with a 409 conflict error. The former one will succeed and the latter one should retry to get the
        existing cluster env.
        """
        existing_cluster_env = self.get_cluster_env_by_name(cluster_env_name)
        if existing_cluster_env is not None:
            return existing_cluster_env

        try:
            if image_uri:
                # For BYOD builds, we should create a build along with the cluster env.
                cluster_environment = self._external_api_client.create_byod_cluster_environment(
                    CreateBYODClusterEnvironment(
                        name=cluster_env_name,
                        config_json=CreateBYODClusterEnvironmentConfigurationSchema(
                            docker_image=image_uri,
                            ray_version=ray_version
                            if ray_version
                            else LATEST_RAY_VERSION,
                            registry_login_secret=registry_login_secret,
                        ),
                        anonymous=anonymous,
                    )
                ).result
            else:
                cluster_environment = self._external_api_client.create_cluster_environment(
                    CreateClusterEnvironment(name=cluster_env_name, anonymous=anonymous)
                ).result
            return cluster_environment
        except ExternalApiException as e:
            if e.status != 409:
                raise e from None
            # Retry to get the existing cluster env because it might be created by another process.
            existing_cluster_env = self.get_cluster_env_by_name(cluster_env_name)
            if existing_cluster_env is None:
                raise e from None
            return existing_cluster_env

    @handle_api_exceptions
    def get_cluster_env_build_id_from_containerfile(
        self,
        cluster_env_name: str,
        containerfile: str,
        anonymous: bool = True,
        ray_version: Optional[str] = None,
    ) -> str:
        cluster_env = self._find_or_create_cluster_env(
            cluster_env_name, anonymous=anonymous
        )
        for build in self.list_cluster_env_builds(cluster_env.id):
            if (
                build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and build.containerfile == containerfile
                # we don't need to check the ray_version because checking the containerfile is enough.
            ):
                return build.id

        try:
            build_op = self._external_api_client.create_cluster_environment_build(
                CreateClusterEnvironmentBuild(
                    cluster_environment_id=cluster_env.id,
                    containerfile=containerfile,
                    ray_version=ray_version,  # we don't use the latest version here if ray_version is Noneb/c the backend will try to parse the base image to decide the ray version.
                )
            ).result
        except ExternalApiException as e:
            if e.status == 400:
                raise RuntimeError(
                    "Invalid containerfile. Please check the syntax and try again.",
                    e.body,
                ) from None

        build_url = self.get_build_ui_url(
            cluster_env.id, build_op.cluster_environment_build_id
        )
        self.logger.info(f"Building image. View it in the UI: {build_url}")
        self._wait_for_build_to_succeed(build_op.cluster_environment_build_id)
        self.logger.info("Image build succeeded.")

        return build_op.cluster_environment_build_id

    @handle_api_exceptions
    def get_cluster_env_build_id_from_image_uri(
        self,
        image_uri: ImageURI,
        registry_login_secret: Optional[str] = None,
        ray_version: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        build = self._internal_api_client.get_or_create_build_from_image_uri_api_v2_builds_get_or_create_build_from_image_uri_post(
            GetOrCreateBuildFromImageUriRequest(
                image_uri=str(image_uri),
                registry_login_secret=registry_login_secret,
                ray_version=ray_version,
                cluster_env_name=name,
            )
        ).result
        return build.id

    @handle_api_exceptions
    def send_workspace_notification(
        self, notification: WorkspaceNotification,
    ):
        if not self.inside_workspace():
            return

        try:
            r = requests.post(WORKSPACE_NOTIFICATION_ADDRESS, json=notification.dict())
            r.raise_for_status()
        except Exception:
            internal_logger.exception(
                "Failed to send workspace notification. "
                "This should not happen, so please contact Anyscale support."
            )

    @handle_api_exceptions
    def get_service(
        self,
        name: str,
        *,
        cloud: Optional[str],
        project: Optional[str],
        include_archived: bool = False,
    ) -> Optional[DecoratedProductionServiceV2APIModel]:
        # we don't have an api to get a service by name, so we need to list services and filter by name
        resp = self.list_services(
            name=name, cloud=cloud, project=project, include_archived=include_archived
        )
        for result in resp.results:
            if result.name == name:
                return result
        return None

    @handle_api_exceptions
    def get_service_by_id(
        self, service_id: str
    ) -> Optional[DecoratedProductionServiceV2APIModel]:
        return self._internal_api_client.get_service_api_v2_services_v2_service_id_get(
            service_id
        ).result

    @handle_api_exceptions
    def list_services(
        self,
        *,
        name: Optional[str] = None,
        state_filter: Optional[List[str]] = None,
        tag_filter: Optional[List[str]] = None,
        creator_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        include_archived: bool = False,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
        sort_field: Optional[str] = None,
        sort_order: Optional[str] = None,
    ) -> DecoratedlistserviceapimodelListResponse:
        cloud_id = self.get_cloud_id(cloud_name=cloud) if cloud else None
        project_id = (
            self.get_project_id(parent_cloud_id=cloud_id, name=project)
            if project
            else None
        )

        return self._internal_api_client.list_services_api_v2_services_v2_get(
            project_id=project_id,
            cloud_id=cloud_id,
            name=name,
            state_filter=state_filter,
            creator_id=creator_id,
            archive_status=ArchiveStatus.ALL
            if include_archived
            else ArchiveStatus.NOT_ARCHIVED,
            tag_filter=tag_filter,
            count=count if count else self.LIST_ENDPOINT_COUNT,
            paging_token=paging_token,
            sort_field=sort_field,
            sort_order=sort_order,
        )

    def get_service_versions(
        self, service_id: str, read_all_versions: bool = False,
    ) -> List[DecoratedProductionServiceV2VersionAPIModel]:
        resp = self._internal_api_client.get_service_versions_api_v2_services_v2_service_id_versions_get(
            service_id=service_id,
        )

        if not read_all_versions:
            return resp.results

        all_versions: List[DecoratedProductionServiceV2VersionAPIModel] = []
        all_versions.extend(resp.results)
        paging_token = resp.metadata.next_paging_token

        while paging_token is not None:
            resp = self._internal_api_client.get_service_versions_api_v2_services_v2_service_id_versions_get(
                service_id=service_id,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
            )
            all_versions.extend(resp.results)
            paging_token = resp.metadata.next_paging_token

        return all_versions

    @handle_api_exceptions
    def get_project(self, project_id: str) -> Project:
        return self._internal_api_client.get_project_api_v2_projects_project_id_get(
            project_id
        ).result

    @handle_api_exceptions
    def list_projects(
        self,
        *,
        name_contains: Optional[str] = None,
        creator_id: Optional[str] = None,
        parent_cloud_id: Optional[str] = None,
        include_defaults: bool = True,
        sort_field: Optional[str] = None,
        sort_order: Optional[str] = None,
        paging_token: Optional[str] = None,
        count: Optional[int] = None,
    ) -> ProjectListResponse:
        return self._internal_api_client.list_projects_api_v2_projects_get(
            name_contains=name_contains,
            creator_id=creator_id,
            parent_cloud_id=parent_cloud_id,
            include_defaults=include_defaults,
            sort_field=sort_field,
            sort_order=sort_order,
            paging_token=paging_token,
            count=count or self.LIST_ENDPOINT_COUNT,
        )

    @handle_api_exceptions
    def create_project(self, project: WriteProject) -> ProjectBase:
        return self._internal_api_client.create_project_api_v2_projects_post(
            project
        ).result

    @handle_api_exceptions
    def delete_project(self, project_id: str) -> None:
        self._internal_api_client.delete_project_api_v2_projects_project_id_delete(
            project_id
        )

    @handle_api_exceptions
    def get_default_project(self, parent_cloud_id: str) -> Project:
        return self._internal_api_client.get_default_project_api_v2_projects_default_project_get(
            parent_cloud_id=parent_cloud_id,
        ).result

    @handle_api_exceptions
    def add_project_collaborators(
        self, project_id: str, collaborators: List[CreateUserProjectCollaborator]
    ) -> None:
        self._internal_api_client.batch_create_project_collaborators_api_v2_projects_project_id_collaborators_users_batch_create_post(
            project_id, collaborators
        )

    @handle_api_exceptions
    def get_job(
        self,
        *,
        name: Optional[str],
        job_id: Optional[str],
        cloud: Optional[str],
        project: Optional[str],
    ) -> Optional[ProductionJob]:
        if job_id is not None:
            try:
                return self._external_api_client.get_production_job(job_id).result
            except ExternalApiException as e:
                if e.status == 404:
                    return None
                raise e from None
        else:
            paging_token = None
            cloud_id = self.get_cloud_id(cloud_name=cloud)
            project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
            result: Optional[ProductionJob] = None
            while True:
                resp = self._external_api_client.list_production_jobs(
                    project_id=project_id,
                    name=name,
                    count=self.LIST_ENDPOINT_COUNT,
                    paging_token=paging_token,
                )
                for job in resp.results:
                    if (
                        job is not None
                        and job.name == name
                        and (result is None or job.created_at > result.created_at)
                    ):
                        result = job

                paging_token = resp.metadata.next_paging_token
                if paging_token is None:
                    break

            return result

    @handle_api_exceptions
    def get_job_runs(self, job_id: str) -> List[APIJobRun]:
        job_runs: List[APIJobRun] = search_entities(
            self._external_api_client.search_jobs,
            JobsQuery(
                ha_job_id=job_id,
                show_ray_client_runs_only=False,
                sort_by_clauses=[
                    SortByClauseJobsSortField(
                        sort_field=JobsSortField.CREATED_AT, sort_order=SortOrder.ASC,
                    )
                ],
                paging=PageQuery(),
            ),
        )
        return job_runs

    @handle_api_exceptions
    def get_job_queue(self, job_queue_id: str) -> Optional[DecoratedJobQueue]:
        try:
            return self._internal_api_client.get_job_queue_api_v2_job_queues_job_queue_id_get(
                job_queue_id
            ).result
        except InternalApiException as e:
            if e.status == 404:
                return None

            raise e from None

    @handle_api_exceptions
    def update_job_queue(
        self,
        job_queue_id: str,
        max_concurrency: Optional[int] = None,
        idle_timeout_s: Optional[int] = None,
    ) -> DecoratedJobQueue:
        if max_concurrency is None and idle_timeout_s is None:
            raise ValueError("No fields to update")

        return self._internal_api_client.update_job_queue_api_v2_job_queues_job_queue_id_put(
            job_queue_id,
            update_job_queue_request=UpdateJobQueueRequest(
                max_concurrency=max_concurrency, idle_timeout_sec=idle_timeout_s,
            ),
        ).result

    @handle_api_exceptions
    def list_job_queues(
        self,
        *,
        name: Optional[str] = None,
        creator_id: Optional[str] = None,
        cluster_status: Optional[SessionState] = None,
        project: Optional[str] = None,
        cloud: Optional[str] = None,
        tags_filter: Optional[Dict[str, List[str]]] = None,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
        sorting_directives: Optional[List[JobQueueSortDirective]] = None,
    ) -> DecoratedjobqueueListResponse:
        cloud_id = self.get_cloud_id(cloud_name=cloud) if cloud else None
        project_id = (
            self.get_project_id(parent_cloud_id=cloud_id, name=project)
            if project
            else None
        )

        return self._internal_api_client.list_job_queues_api_v2_job_queues_post(
            job_queues_query=JobQueuesQuery(
                name=TextQuery(equals=name) if name else None,
                creator_id=creator_id,
                cluster_status=cluster_status,
                project_id=project_id,
                cloud_id=cloud_id,
                tags_filter=tags_filter,
                paging=PageQuery(count=count, paging_token=paging_token),
                sorting_directives=sorting_directives,
            ),
        )

    @handle_api_exceptions
    def list_jobs(
        self,
        *,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        creator_id: Optional[str] = None,
        state_filter: Optional[List[str]] = None,
        archive_status: Optional[str] = None,
        tags_filter: Optional[List[str]] = None,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
    ) -> DecoratedproductionjobListResponse:
        # Build kwargs dynamically to avoid passing None for count,
        # which causes TypeError in OpenAPI client validation
        kwargs: Dict[str, Any] = {
            "project_id": project_id,
            "name": name,
            "creator_id": creator_id,
            "type_filter": "BATCH_JOB",
            "archive_status": archive_status,
            "state_filter": state_filter,
            "tag_filter": tags_filter,
            "paging_token": paging_token,
        }
        if count is not None:
            kwargs["count"] = count
        return self._internal_api_client.list_decorated_jobs_api_v2_decorated_ha_jobs_get(
            **kwargs
        )

    @handle_api_exceptions
    def rollout_service(
        self, model: ApplyProductionServiceV2Model
    ) -> DecoratedProductionServiceV2APIModel:
        result = self._internal_api_client.apply_service_api_v2_services_v2_apply_put(
            model
        ).result
        return result

    @handle_api_exceptions
    def rollout_service_multi_version(
        self, model: ApplyProductionServiceMultiVersionV2Model
    ) -> DecoratedProductionServiceV2APIModel:
        result = self._internal_api_client.apply_service_multi_version_api_v2_services_v2_apply_multi_version_put(
            model
        ).result
        return result

    @handle_api_exceptions
    def rollback_service(
        self, service_id: str, *, max_surge_percent: Optional[int] = None
    ) -> DecoratedProductionServiceV2APIModel:
        result = self._internal_api_client.rollback_service_api_v2_services_v2_service_id_rollback_post(
            service_id,
            rollback_service_model=RollbackServiceModel(
                max_surge_percent=max_surge_percent
            ),
        )
        return result

    @handle_api_exceptions
    def terminate_service(
        self, service_id: str
    ) -> DecoratedProductionServiceV2APIModel:
        result = self._internal_api_client.terminate_service_api_v2_services_v2_service_id_terminate_post(
            service_id
        )
        return result

    @handle_api_exceptions
    def archive_service(self, service_id: str) -> DecoratedProductionServiceV2APIModel:
        result = self._internal_api_client.archive_service_api_v2_services_v2_service_id_archive_post(
            service_id
        )
        return result

    @handle_api_exceptions
    def delete_service(self, service_id: str) -> None:
        self._internal_api_client.delete_service_api_v2_services_v2_service_id_delete(
            service_id
        )

    @handle_api_exceptions
    def submit_job(self, model: CreateInternalProductionJob) -> InternalProductionJob:
        job: InternalProductionJob = self._internal_api_client.create_job_api_v2_decorated_ha_jobs_create_post(
            model,
        ).result
        return job

    @handle_api_exceptions
    def terminate_job(self, job_id: str):
        self._external_api_client.terminate_job(job_id)

    @handle_api_exceptions
    def archive_job(self, job_id: str):
        self._internal_api_client.archive_job_api_v2_decorated_ha_jobs_production_job_id_archive_post(
            job_id
        )

    def _upload_local_runtime_env(
        self,
        cloud_id: str,
        cloud_resource_id: Optional[str],
        zip_file_bytes: bytes,
        content_hash: str,
        overwrite_existing_file: bool,
    ) -> CloudDataBucketPresignedUrlResponse:
        file_name = RUNTIME_ENV_PACKAGE_FORMAT.format(content_hash=content_hash)
        request = CloudDataBucketPresignedUrlRequest(
            file_type=CloudDataBucketFileType.RUNTIME_ENV_PACKAGES,
            file_name=file_name,
            access_mode=CloudDataBucketAccessMode.WRITE,
            cloud_resource_id=cloud_resource_id,
        )
        info: CloudDataBucketPresignedUrlResponse = self._internal_api_client.generate_cloud_data_bucket_presigned_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_url_post(
            cloud_id, request
        ).result

        # Skip the upload entirely if the file already exists.
        if info.file_exists and not overwrite_existing_file:
            internal_logger.debug(
                f"Skipping file upload for '{file_name}' because it already exists in cloud storage."
            )
            return info

        if info.url_scheme == CloudDataBucketPresignedUrlScheme.SMART_OPEN:
            # If the presigned URL scheme is SMART_OPEN, upload to cloud storage using the provided bucket name, path, & environment, and the smart_open library.
            bucket_name = info.bucket_name
            bucket_path = info.bucket_path

            env_vars: Dict[str, str] = {
                "AWS_ENDPOINT_URL": info.url,
            }
            with set_env(**env_vars), smart_open.open(
                f"{bucket_name}/{bucket_path}", "wb",
            ) as fout:
                fout.write(zip_file_bytes)

        else:
            # Default to HTTP PUT.
            internal_logger.debug(f"Uploading file '{file_name}' to cloud storage.")
            headers = None
            if info.file_uri.startswith("azure") or info.file_uri.startswith("abfss"):
                headers = {
                    "x-ms-blob-type": "BlockBlob",
                    "x-ms-version": "2025-07-05",
                    "x-ms-date": datetime.utcnow().strftime(
                        "%a, %d %b %Y %H:%M:%S GMT"
                    ),
                    "x-ms-blob-content-type": "application/zip",
                }
            requests.put(
                info.url, data=zip_file_bytes, headers=headers
            ).raise_for_status()

        return info

    @handle_api_exceptions
    def upload_local_dir_to_cloud_storage(
        self,
        local_dir: str,
        *,
        cloud_id: str,
        excludes: Optional[List[str]] = None,
        overwrite_existing_file: bool = OVERWRITE_EXISTING_CLOUD_STORAGE_FILES,
        cloud_resource_name: Optional[str] = None,
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        cloud_resource_id = None
        if cloud_resource_name is not None:
            cloud_resources = self._internal_api_client.get_cloud_resources_api_v2_clouds_cloud_id_resources_get(
                cloud_id=cloud_id,
            ).results
            cloud_resource_id = next(
                (
                    cloud_resource.cloud_resource_id
                    for cloud_resource in cloud_resources
                    if cloud_resource.name == cloud_resource_name
                ),
                None,
            )
            if cloud_resource_id is None:
                raise ValueError(
                    f"Cloud resource '{cloud_resource_name}' not found in cloud '{cloud_id}'"
                )

        with zip_local_dir(local_dir, excludes=excludes) as (
            _,
            zip_file_bytes,
            content_hash,
        ):
            info = self._upload_local_runtime_env(
                cloud_id=cloud_id,
                cloud_resource_id=cloud_resource_id,
                zip_file_bytes=zip_file_bytes,
                content_hash=content_hash,
                overwrite_existing_file=overwrite_existing_file,
            )
            return info.file_uri

    def upload_local_dir_to_cloud_storage_multi_cloud_resource(
        self,
        local_dir: str,
        *,
        cloud_id: str,
        cloud_resource_names: List[Optional[str]],
        excludes: Optional[List[str]] = None,
        overwrite_existing_file: bool = False,
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        all_cloud_resources = self._internal_api_client.get_cloud_resources_api_v2_clouds_cloud_id_resources_get(
            cloud_id=cloud_id,
        ).results
        cloud_resource_names_to_ids = {
            cloud_resource.name: cloud_resource.cloud_resource_id
            for cloud_resource in all_cloud_resources
        }

        bucket_paths = set()

        with zip_local_dir(local_dir, excludes=excludes) as (
            _,
            zip_file_bytes,
            content_hash,
        ):
            for cloud_resource_name in cloud_resource_names:
                if cloud_resource_name is not None:
                    if cloud_resource_name not in cloud_resource_names_to_ids:
                        raise ValueError(
                            f"Cloud resource '{cloud_resource_name}' not found in cloud '{cloud_id}'"
                        )
                    cloud_resource_id = cloud_resource_names_to_ids[cloud_resource_name]
                else:
                    cloud_resource_id = None

                info = self._upload_local_runtime_env(
                    cloud_id=cloud_id,
                    cloud_resource_id=cloud_resource_id,
                    zip_file_bytes=zip_file_bytes,
                    content_hash=content_hash,
                    overwrite_existing_file=overwrite_existing_file,
                )
                bucket_paths.add(info.bucket_path)

            assert len(bucket_paths) == 1
            return bucket_paths.pop()

    def _fetch_log_chunks(self, job_run_id: str) -> Tuple[List[str], Any]:
        all_log_chunk_urls = []
        MAX_PAGE_SIZE = 1000
        next_page_token = None
        bearer_token = None
        while True:
            log_download_result = self._internal_api_client.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
                job_id=job_run_id,
                next_page_token=next_page_token,
                page_size=MAX_PAGE_SIZE,
            ).result

            if bearer_token is None:
                bearer_token = log_download_result.bearer_token

            all_log_chunk_urls.extend(
                [chunk.chunk_url for chunk in log_download_result.log_chunks]
            )

            next_page_token = log_download_result.next_page_token
            if next_page_token is None:
                break

        return all_log_chunk_urls, bearer_token

    def _fetch_log_chunks_for_controller_logs(
        self, cluster_id: str
    ) -> Tuple[List[str], Any]:
        all_log_chunk_urls = []
        MAX_PAGE_SIZE = 1000
        next_page_token = None
        bearer_token = None
        while True:
            log_download_result = self._internal_api_client.get_serve_logs_download_api_v2_logs_serve_logs_download_cluster_id_get(
                cluster_id=cluster_id,
                next_page_token=next_page_token,
                page_size=MAX_PAGE_SIZE,
            ).result

            if bearer_token is None:
                bearer_token = log_download_result.bearer_token

            all_log_chunk_urls.extend(
                [chunk.chunk_url for chunk in log_download_result.log_chunks]
            )

            next_page_token = log_download_result.next_page_token
            if next_page_token is None:
                break

        return all_log_chunk_urls, bearer_token

    def _read_log_lines(  # noqa: PLR0912
        self,
        log_chunk_urls: List[str],
        head: bool,
        bearer_token: Any,
        max_lines: Optional[int],
        parse_json: Optional[bool] = None,
    ) -> str:
        # TODO(mowen): Would be nice to return some placeholder here to allow all new
        # logs to be read from a particular point onwards.
        # TODO(aguo): Change this to be a generator to avoid loading all logs into memory at once
        # and to gradually load logs as needed.

        def parse_json_line(line: str) -> str:
            json_line = json.loads(line)
            # This is the default schema for ray core logger but users
            # could technically use any schema they want for structured logs.
            # Fall back to spitting out the json in the worst-case scenario.
            if "asctime" in json_line and "message" in json_line:
                return f"{json_line['asctime']} {json_line['message']}"
            elif "message" in json_line:
                return json_line["message"]
            # This is the worst-case scenario but very unlikely. Users would
            # not likely be changing "message" to something-else.
            return line

        result_lines: List[str] = []
        step = 1 if head else -1
        line_count = 0
        for chunk_url in log_chunk_urls[::step]:
            log_lines = _download_log_from_s3_url_sync(
                chunk_url, bearer_token=bearer_token
            ).splitlines()

            if max_lines is not None:
                num_lines_to_add = min(len(log_lines), max_lines - line_count)
            else:
                num_lines_to_add = len(log_lines)
            line_count += num_lines_to_add

            if head:
                lines_to_add = log_lines[:num_lines_to_add]
            else:
                lines_to_add = log_lines[-1 * num_lines_to_add :]

            if parse_json is not False:
                try:
                    lines_to_add = [parse_json_line(line) for line in lines_to_add]
                except json.JSONDecodeError:
                    if parse_json is True:
                        raise ValueError(
                            "Failed to parse logs as JSON. Logs are not all in JSON format."
                        )
                    # If we fail to parse_json, we should always just use plain text going forward.
                    parse_json = False
                    # lines_to_add should already be plain text, so continue on...

            result_lines = (
                result_lines + lines_to_add if head else lines_to_add + result_lines
            )

            if line_count == max_lines:
                break

        if not result_lines:
            return ""
        return "\n".join(result_lines) + "\n"

    @handle_api_exceptions
    def logs_for_job_run(
        self,
        job_run_id: str,
        head: bool = False,
        max_lines: Optional[int] = None,
        parse_json: Optional[bool] = None,
    ) -> str:
        """
        Retrieves logs from the streaming job logs folder in S3/GCS
        Args:
        - parse_json: If true, we will always attempt to parse the logs as JSON.
            If false, we will always attempt to parse the logs as text. If None, we
            will attempt to parse the logs as JSON and fall back to text if parsing
            fails.
        """

        all_log_chunk_urls, bearer_token = self._fetch_log_chunks(job_run_id)

        logs = self._read_log_lines(
            all_log_chunk_urls, head, bearer_token, max_lines, parse_json=parse_json
        )
        return logs

    @handle_api_exceptions
    def stream_logs_for_job_run(
        self, job_run_id: str, next_page_token: Optional[str] = None,
    ) -> Tuple[str, Optional[str]]:
        """Stream logs incrementally for a job run with pagination support.

        Args:
            job_run_id: The ID of the job run to fetch logs for
            next_page_token: Token for fetching the next page of logs (for incremental streaming)

        Returns:
            Tuple of (logs, next_page_token) where next_page_token can be used for the next call
        """
        # Fetch only the new log chunks since the last call
        if next_page_token:
            # Incremental fetch - get only new chunks
            log_download_result = self._internal_api_client.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
                job_id=job_run_id, next_page_token=next_page_token,
            ).result
        else:
            # First fetch - get all available chunks
            log_download_result = self._internal_api_client.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
                job_id=job_run_id,
            ).result

        # Download and concatenate log chunks
        log_chunk_urls = [chunk.chunk_url for chunk in log_download_result.log_chunks]
        bearer_token = log_download_result.bearer_token

        logs = self._read_log_lines(
            log_chunk_urls,
            head=False,
            bearer_token=bearer_token,
            max_lines=None,
            parse_json=False,
        )

        # Return logs and the token for the next page
        new_next_page_token = (
            log_download_result.next_page_token
            if len(log_download_result.log_chunks) > 0
            else next_page_token
        )

        return logs, new_next_page_token

    @handle_api_exceptions
    def controller_logs_for_service_version(
        self,
        service_version: ProductionServiceV2VersionModel,
        head: bool = False,
        max_lines: Optional[int] = None,
        parse_json: Optional[bool] = None,
    ) -> str:
        """
        Returns the controller logs associated with a particular service version.

        Args:
        - parse_json: If true, we will always attempt to parse the logs as JSON.
            If false, we will always attempt to parse the logs as text. If None, we
            will attempt to parse the logs as JSON and fall back to text if parsing
            fails.
        """
        service_version_id = service_version.id[-8:]
        if not len(service_version.production_job_ids):
            raise ValueError(
                f"Service version '{service_version_id}' canary version is not ready. Please try again later..."
            )

        job = self._internal_api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
            service_version.production_job_ids[0]
        ).result

        if not job:
            raise ValueError(
                f"Service version '{service_version_id}' canary version is not ready. Please try again later..."
            )

        if not job.state.cluster:
            raise ValueError(
                f"Service version '{service_version_id}' canary version is not ready. Please try again later..."
            )

        cluster_id = job.state.cluster.id

        all_log_chunk_urls, bearer_token = self._fetch_log_chunks_for_controller_logs(
            cluster_id
        )

        logs = self._read_log_lines(
            all_log_chunk_urls, head, bearer_token, max_lines, parse_json=parse_json
        )
        return logs

    @handle_api_exceptions
    def apply_schedule(self, model: CreateSchedule) -> DecoratedSchedule:
        return self._internal_api_client.create_or_update_job_api_v2_experimental_cron_jobs_put(
            model
        ).result

    @handle_api_exceptions
    def get_schedule(
        self,
        *,
        name: Optional[str],
        id: Optional[str],  # noqa: A002
        cloud: Optional[str],
        project: Optional[str],
    ) -> Optional[DecoratedSchedule]:
        if id is not None:
            try:
                return self._internal_api_client.get_cron_job_api_v2_experimental_cron_jobs_cron_job_id_get(
                    id
                ).result
            except ExternalApiException as e:
                if e.status == 404:
                    return None
                raise e from None
        else:
            paging_token = None
            cloud_id = self.get_cloud_id(cloud_name=cloud)
            project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
            result: Optional[DecoratedSchedule] = None
            while True:
                resp = self._internal_api_client.list_cron_jobs_api_v2_experimental_cron_jobs_get(
                    project_id=project_id,
                    name=name,
                    count=self.LIST_ENDPOINT_COUNT,
                    paging_token=paging_token,
                )
                for schedule in resp.results:
                    if schedule is not None and schedule.name == name:
                        result = schedule
                        break

                paging_token = resp.metadata.next_paging_token
                if paging_token is None:
                    break

            return result

    @handle_api_exceptions
    def set_schedule_state(self, id: str, is_paused: bool):  # noqa: A002
        _ = self._internal_api_client.pause_cron_job_api_v2_experimental_cron_jobs_cron_job_id_pause_post(
            id, {"is_paused": is_paused}
        ).result

    @handle_api_exceptions
    def trigger_schedule(self, id: str):  # noqa: A002
        self._internal_api_client.trigger_cron_job_api_v2_experimental_cron_jobs_cron_job_id_trigger_post(
            id
        )

    @handle_api_exceptions
    def list_schedules(
        self,
        *,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        cloud_id: Optional[str] = None,
        creator_id: Optional[str] = None,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
    ) -> DecoratedscheduleListResponse:
        """List schedules with filtering and pagination."""
        # Build kwargs dynamically to avoid passing None for count,
        # which causes TypeError in OpenAPI client validation
        kwargs: Dict[str, Any] = {
            "project_id": project_id,
            "cloud_id": cloud_id,
            "name": name,
            "creator_id": creator_id,
            "paging_token": paging_token,
        }
        if count is not None:
            kwargs["count"] = count
        return self._internal_api_client.list_cron_jobs_api_v2_experimental_cron_jobs_get(
            **kwargs
        )

    @handle_api_exceptions
    def create_workspace(self, model: CreateExperimentalWorkspace) -> str:
        return self._internal_api_client.create_workspace_api_v2_experimental_workspaces_post(
            create_experimental_workspace=model,
        ).result.id

    @handle_api_exceptions
    def get_workspace(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> Optional[ExperimentalWorkspace]:
        """Get a workspace by either name or id. Filter by cloud and project.

        Returns None if not found.
        """
        if id is not None:
            try:
                return self._internal_api_client.get_workspace_api_v2_experimental_workspaces_workspace_id_get(
                    id
                ).result
            except ExternalApiException as e:
                if e.status == 404:
                    return None
                raise e from None
        else:
            paging_token = None
            cloud_id = self.get_cloud_id(cloud_name=cloud)
            project_id = self.get_project_id(parent_cloud_id=cloud_id, name=project)
            resp = self._internal_api_client.list_workspaces_api_v2_experimental_workspaces_get(
                project_id=project_id,
                name=name,
                count=self.LIST_ENDPOINT_COUNT,
                paging_token=paging_token,
            )

            if len(resp.results) == 0:
                return None

            workspace = resp.results[0]
            return workspace

    @handle_api_exceptions
    def update_workspace(
        self,
        *,
        workspace_id: Optional[str],
        name: Optional[str] = None,
        compute_config_id: Optional[str] = None,
        cluster_environment_build_id: Optional[str] = None,
        idle_timeout_minutes: Optional[int] = None,
    ):
        workspace = self.get_workspace(id=workspace_id)
        if not workspace:
            raise ValueError(f"Workspace '{workspace_id}' not found.")

        if name:
            # Update the workspace name with workspaces patch API
            self._internal_api_client.patch_workspace_api_v2_experimental_workspaces_workspace_id_patch(
                workspace_id=workspace_id,
                json_patch_operation=[
                    {"op": "replace", "path": "/name", "value": name,},
                ],
            )

        if compute_config_id or cluster_environment_build_id or idle_timeout_minutes:
            # Use the internal session API to update cluster config and idle timeout
            # This matches what the UI does and is more efficient than the external API
            self._internal_api_client.put_session_cluster_config_with_session_idle_timeout_api_v2_sessions_session_id_cluster_config_with_session_idle_timeout_put(
                session_id=workspace.cluster_id,
                compute_template_id=compute_config_id,
                build_id=cluster_environment_build_id,
                idle_timeout=idle_timeout_minutes,
            )

    @handle_api_exceptions
    def update_workspace_dependencies_offline_only(
        self, workspace_id: str, requirements: List[str]
    ):
        return self._internal_api_client.put_workspace_proxied_dataplane_artifacts_api_v2_experimental_workspaces_workspace_id_proxied_dataplane_artifacts_put(
            workspace_id=workspace_id,
            workspace_dataplane_proxied_artifacts={
                "requirements": "\n".join(requirements),
            },
        )

    @handle_api_exceptions
    def update_workspace_env_vars_offline_only(
        self, workspace_id: str, env_vars: Dict[str, str]
    ):
        return self._internal_api_client.put_workspace_proxied_dataplane_artifacts_api_v2_experimental_workspaces_workspace_id_proxied_dataplane_artifacts_put(
            workspace_id=workspace_id,
            workspace_dataplane_proxied_artifacts={
                "environment_variables": [
                    f"{key}={value}" for key, value in env_vars.items()
                ],
            },
        )

    @handle_api_exceptions
    def get_workspace_proxied_dataplane_artifacts(
        self, workspace_id: str
    ) -> WorkspaceDataplaneProxiedArtifacts:
        return self._internal_api_client.get_workspace_proxied_dataplane_artifacts_api_v2_experimental_workspaces_workspace_id_proxied_dataplane_artifacts_get(
            workspace_id
        ).result

    @handle_api_exceptions
    def start_workspace(self, workspace_id: str):
        """Start a workspace."""
        workspace_model = self.get_workspace(id=workspace_id)
        if workspace_model is None:
            raise ValueError(f"Workspace '{workspace_id}' not found.")

        return self._internal_api_client.start_session_api_v2_sessions_session_id_start_post(
            session_id=workspace_model.cluster_id,
            start_session_options=StartSessionOptions(),
        )

    @handle_api_exceptions
    def terminate_workspace(self, workspace_id: str):
        """Terminate a workspace."""
        workspace_model = self.get_workspace(id=workspace_id)
        if workspace_model is None:
            raise ValueError(f"Workspace '{workspace_id}' not found.")

        options = StopSessionOptions(
            terminate=True,
            workers_only=False,
            keep_min_workers=False,
            take_snapshot=False,
        )
        return self._internal_api_client.stop_session_api_v2_sessions_session_id_stop_post(
            session_id=workspace_model.cluster_id, stop_session_options=options,
        )

    @handle_api_exceptions
    def get_workspace_cluster(
        self, workspace_id: Optional[str]
    ) -> Optional[DecoratedSession]:
        workspace = self.get_workspace(id=workspace_id)
        if not workspace:
            raise ValueError(f"Workspace '{workspace_id}' not found.")

        result = self._internal_api_client.get_decorated_cluster_api_v2_decorated_sessions_cluster_id_get(
            workspace.cluster_id
        )
        return result.result

    @handle_api_exceptions
    def get_cluster_head_node_ip(self, cluster_id: str) -> str:
        head_ip = self._internal_api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
            cluster_id
        ).result.head_ip
        return head_ip

    @handle_api_exceptions
    def get_cluster_ssh_key(self, cluster_id: str) -> SessionSshKey:
        return self._internal_api_client.get_session_ssh_key_api_v2_sessions_session_id_ssh_key_get(
            cluster_id
        ).result

    @handle_api_exceptions
    def get_workspace_default_dir_name(self, workspace_id) -> str:
        workspace = self.get_workspace(id=workspace_id)
        assert workspace, f"Workspace '{workspace_id}' not found."
        project = self._internal_api_client.get_project_api_v2_projects_project_id_get(
            workspace.project_id
        ).result
        if self._internal_api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
            FLAG_DEFAULT_WORKING_DIR_FOR_PROJ
        ).result.is_on:
            return project.directory_name
        else:
            return project.name

    @handle_api_exceptions
    def download_aggregated_instance_usage_csv(
        self,
        start_date,
        end_date,
        cloud_id=None,
        project_id=None,
        directory=None,
        hide_progress_bar=False,
    ) -> str:
        with FileDownloadProgress() as progress:
            task_id = progress.add_task(
                description="Preparing aggregated instance usage CSV",
                visible=not hide_progress_bar,
            )

            resp = self._internal_api_client.download_aggregated_instance_usage_csv_api_v2_aggregated_instance_usage_download_csv_get(
                start_date=start_date,
                end_date=end_date,
                cloud_id=cloud_id,
                project_id=project_id,
                _preload_content=False,
            )

            progress.update(
                task_id, description="Downloading aggregated instance usage CSV"
            )

            # Set the total size of the download for the progress bar
            total_size = int(resp.headers.get("content-length", 0))
            progress.update(task_id, total=total_size)

            # Construct the filepath to save the downloaded file
            content_disposition = resp.headers.get("content-disposition", "")
            filename_regex = re.search(
                r'filename="?(?P<filename>[^"]+)"?', content_disposition
            )

            if not filename_regex:
                filename = f"aggregated_instance_usage_{start_date}_{end_date}.zip"
            else:
                filename = filename_regex.group("filename")

            filepath = os.path.join(directory, filename) if directory else filename

            # Download the file
            try:
                with open(filepath, "wb") as f:
                    for chunk in resp.stream(Bytes.MB):
                        if chunk:
                            f.write(chunk)
                            progress.update(task_id, advance=len(chunk))
            except Exception as e:  # noqa: BLE001
                raise ValueError(f"Failed to download to '{filepath}': {e}") from None

            progress.update(task_id, completed=total_size)
            progress.console.print(
                f"Download complete! File saved to '{filepath}'",
                style=Style(bold=True, color="green"),
            )

        return filepath

    @handle_api_exceptions
    def create_api_key(
        self, duration: float, user_id: Optional[str]
    ) -> ServerSessionToken:
        return self._internal_api_client.create_api_key_api_v2_users_create_api_key_post(
            ApiKeyParameters(user_id=user_id, duration=duration)
        ).result

    @handle_api_exceptions
    def rotate_api_key(self, user_id: str) -> None:
        self._internal_api_client.rotate_api_key_for_user_api_v2_organization_collaborators_rotate_api_key_for_user_user_id_post(
            user_id
        )

    @handle_api_exceptions
    def admin_batch_create_users(
        self, admin_create_users: List[AdminCreateUser]
    ) -> List[AdminCreatedUser]:
        return self._internal_api_client.admin_batch_create_users_api_v2_users_admin_batch_create_post(
            admin_create_users
        ).results

    @handle_api_exceptions
    def create_organization_invitations(
        self, emails: List[str]
    ) -> Tuple[List[str], List[str]]:
        results = self._internal_api_client.batch_create_invitations_api_v2_organization_invitations_batch_create_post(
            [CreateOrganizationInvitation(email=email) for email in emails]
        ).results

        success_emails = []
        error_messages = []

        for idx, result in enumerate(results):
            if result.data:
                success_emails.append(emails[idx])
            else:
                error_messages.append(result.error.detail)

        return success_emails, error_messages

    @handle_api_exceptions
    def list_organization_invitations(self) -> List[OrganizationInvitation]:
        results = (
            self._internal_api_client.list_invitations_api_v2_organization_invitations_get().results
        )

        return results

    @handle_api_exceptions
    def delete_organization_invitation(self, email: str) -> OrganizationInvitation:
        invitation = self._internal_api_client.list_invitations_api_v2_organization_invitations_get(
            email=email
        ).results

        if len(invitation) == 0:
            raise ValueError(f"Invitation for email '{email}' not found.")
        elif len(invitation) > 1:
            raise ValueError(
                f"Multiple invitations found for email '{email}'. Please contact Anyscale support."
            )

        invitation_id = invitation[0].id

        return self._internal_api_client.invalidate_invitation_api_v2_organization_invitations_invitation_id_invalidate_post(
            invitation_id
        ).result

    @handle_api_exceptions
    def list_organization_collaborators(
        self,
        *,
        email: Optional[str] = None,
        name: Optional[str] = None,
        collaborator_type: Optional[CollaboratorType] = None,
        is_service_account: Optional[bool] = None,
        count: Optional[int] = None,
        paging_token: Optional[str] = None,
    ) -> OrganizationcollaboratorListResponse:
        collaborator_type_value = None
        if collaborator_type is not None:
            collaborator_type_value = (
                collaborator_type.value
                if hasattr(collaborator_type, "value")
                else str(collaborator_type)
            )

        return self._internal_api_client.list_organization_collaborators_api_v2_organization_collaborators_get(
            email=email,
            name=name,
            collaborator_type=collaborator_type_value,
            is_service_account=is_service_account,
            count=count if count is not None else self.LIST_ENDPOINT_COUNT,
            paging_token=paging_token,
        )

    @handle_api_exceptions
    def get_organization_collaborators(
        self,
        email: Optional[str] = None,
        name: Optional[str] = None,
        collaborator_type: Optional[CollaboratorType] = None,
        is_service_account: Optional[bool] = None,  # noqa: ARG002
    ) -> List[OrganizationCollaborator]:
        response = self.list_organization_collaborators(
            email=email,
            name=name,
            collaborator_type=collaborator_type,
            is_service_account=is_service_account,
        )

        return response.results

    @handle_api_exceptions
    def delete_organization_collaborator(self, identity_id: str) -> None:
        self._internal_api_client.remove_organization_collaborator_api_v2_organization_collaborators_identity_id_delete(
            identity_id
        )

    @handle_api_exceptions
    def create_service_account(self, name: str) -> AnyscaleServiceAccount:
        return self._internal_api_client.create_service_account_api_v2_users_service_accounts_post(
            name
        ).result

    @handle_api_exceptions
    def create_resource_quota(
        self, create_resource_quota: CreateResourceQuota
    ) -> ResourceQuota:
        return self._internal_api_client.create_resource_quota_api_v2_resource_quotas_post(
            create_resource_quota
        ).result

    @handle_api_exceptions
    def list_resource_quotas(
        self,
        name: Optional[str] = None,
        cloud_id: Optional[str] = None,
        creator_id: Optional[str] = None,
        is_enabled: Optional[bool] = None,
        max_items: int = 20,
    ) -> List[ResourceQuota]:

        query = ListResourceQuotasQuery(
            name=TextQuery(equals=name) if name else None,
            cloud_id=cloud_id,
            creator_id=creator_id,
            is_enabled=is_enabled,
            paging=PageQuery(count=max_items),
        )

        resource_quotas = self._internal_api_client.search_resource_quotas_api_v2_resource_quotas_search_post(
            query
        ).results

        return resource_quotas

    @handle_api_exceptions
    def delete_resource_quota(self, resource_quota_id: str) -> None:
        self._internal_api_client.delete_resource_quota_api_v2_resource_quotas_resource_quota_id_delete(
            resource_quota_id
        )

    @handle_api_exceptions
    def set_resource_quota_status(
        self, resource_quota_id: str, is_enabled: bool
    ) -> None:
        _ = self._internal_api_client.set_resource_quota_status_api_v2_resource_quotas_resource_quota_id_status_patch(
            resource_quota_id, ResourceQuotaStatus(is_enabled=is_enabled)
        ).result

    @handle_api_exceptions
    def upsert_resource_tags(
        self,
        resource_type: ResourceTagResourceType,
        resource_id: str,
        tags: Dict[str, str],
    ) -> None:
        req = UpsertResourceTagsRequest(
            resource_type=resource_type, resource_id=resource_id, tags=tags
        )
        self._internal_api_client.upsert_resource_tags_api_v2_tags_resource_put(req)

    @handle_api_exceptions
    def delete_resource_tags(
        self, resource_type: ResourceTagResourceType, resource_id: str, keys: List[str],
    ) -> None:
        req = DeleteResourceTagsRequest(
            resource_type=resource_type, resource_id=resource_id, keys=keys
        )
        self._internal_api_client.delete_resource_tags_api_v2_tags_resource_delete(req)

    @handle_api_exceptions
    def list_resource_tags(
        self, resource_type: ResourceTagResourceType, resource_id: str
    ) -> List[ResourceTagRecord]:
        resp = self._internal_api_client.get_tags_for_resource_api_v2_tags_resource_get(
            resource_type, resource_id
        )
        result = resp.result
        if result is None or result.tags is None:
            return []
        return list(result.tags)

    @handle_api_exceptions
    def list_user_groups(
        self, *, count: int = 50, paging_token: Optional[str] = None,
    ) -> UsergroupListResponse:
        return self._internal_api_client.list_user_groups_api_v2_user_groups_get(
            count=count, paging_token=paging_token,
        )

    @handle_api_exceptions
    def get_user_group(self, group_id: str) -> UserGroup:
        response = self._internal_api_client.get_user_group_api_v2_user_groups_group_id_get(
            group_id
        )
        return response.result

    @handle_api_exceptions
    def list_user_group_memberships(self) -> Dict:
        """List all user groups with their members."""
        api = self._internal_api_client
        response = api.api_client.call_api(
            "/api/v2/user_groups/memberships/list",
            "GET",
            query_params=[],
            response_type="object",
            auth_settings=["HTTPBearer"],
            _return_http_data_only=True,
        )
        return response

    @handle_api_exceptions
    def update_resource_policy(
        self, resource_type: str, resource_id: str, policy: UpdatePolicyRequest,
    ) -> None:
        self._internal_api_client.update_resource_policy_api_v2_policy_resource_type_resource_id_put(
            resource_type=resource_type,
            resource_id=resource_id,
            update_policy_request=policy,
        )

    @handle_api_exceptions
    def get_resource_policy(
        self, resource_type: str, resource_id: str,
    ) -> PolicyResponse:
        response = self._internal_api_client.get_resource_policy_api_v2_policy_resource_type_resource_id_get(
            resource_type=resource_type, resource_id=resource_id,
        )
        return response.result

    @handle_api_exceptions
    def list_resource_policies(
        self, resource_type: str,
    ) -> ResourcepolicyitemListResponse:
        return self._internal_api_client.list_resource_policies_api_v2_policy_resource_type_get(
            resource_type=resource_type,
        )

    @handle_api_exceptions
    def migrate_scim_permissions(self, *, dry_run: bool = True) -> Dict:
        """Migrate organization permissions to SCIM-based user group permissions."""
        api = self._internal_api_client
        if hasattr(
            api, "migrate_scim_permissions_api_v2_scim_migrate_permissions_post"
        ):
            resp = api.migrate_scim_permissions_api_v2_scim_migrate_permissions_post(
                dry_run=dry_run
            )
            if hasattr(resp, "to_dict"):
                return resp.to_dict()  # type: ignore[no-any-return]
            return {"result": getattr(resp, "result", resp)}

        response = api.api_client.call_api(
            "/api/v2/scim/migrate-permissions",
            "POST",
            query_params=[("dry_run", dry_run)],
            response_type="object",
            auth_settings=["HTTPBearer"],
            _return_http_data_only=True,
        )
        return response

    @handle_api_exceptions
    def list_scim_user_permissions(self, user_id: Optional[str] = None) -> Dict:
        """List users and their effective cloud/project permissions, plus org owners."""
        api = self._internal_api_client
        query_params = []
        if user_id:
            query_params.append(("user_id", user_id))
        response = api.api_client.call_api(
            "/api/v2/scim/list-user-permissions",
            "GET",
            query_params=query_params,
            response_type="object",
            auth_settings=["HTTPBearer"],
            _return_http_data_only=True,
        )
        return response

    @handle_api_exceptions
    def get_user_info(self) -> UserInfo:
        """Get information about the current user."""
        return self._internal_api_client.get_user_info_api_v2_userinfo_get().result
